/* -*- mode: C; mode: fold; -*- */
/*
Copyright (C) 2009 John E. Davis

This file is part of the S-Lang mpfp Module

The S-Lang mpfp Module is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The S-Lang mpfi Module is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.
*/

#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#include <slang.h>

#include <mpfr.h>
#include <mpfi.h>
#include <mpfi_io.h>

#ifdef __cplusplus
extern "C"
{
#endif
SLANG_MODULE(mpfp);
#ifdef __cplusplus
}
#endif

#include "version.h"

#if SLANG_VERSION < 20201
SL_EXTERN SLang_Array_Type *SLang_create_array1 (SLtype, int, VOID_STAR, SLindex_Type *, unsigned int, int);
#endif

/* Memory allocation functions */ /*{{{*/

/* Note concerning malloc failures.  The source code and documentation for
 * gmp, mpfr, and mpfi reveals that no error handling is done: all mallocs
 * are assumed to succeed.  About the best that one can do is to install
 * a custom error handler and then longjmp.  Of course that is ugly and may
 * produce memory leaks and undefined behavior.
 */
static jmp_buf LongJmpEnv;
static void *mem_malloc (size_t len)
{
   char *ptr;

   if (NULL == (ptr = SLmalloc (len)))
     longjmp (LongJmpEnv, 1);

   return ptr;
}

static void *mem_realloc (void *ptr, size_t old_size, size_t new_size)
{
   void *newptr;
   (void) old_size;
   if (NULL == (newptr = SLrealloc ((SLFUTURE_VOID *)ptr, new_size)))
     longjmp (LongJmpEnv, 1);

   return newptr;
}

static void mem_free (void *ptr, size_t size)
{
   (void) size;
   SLfree ((char *) ptr);
}

/*}}}*/

#define NUM_INT_TYPES 4
static SLtype Integer_Types[NUM_INT_TYPES] =
{
   SLANG_CHAR_TYPE,
   SLANG_SHORT_TYPE,
   SLANG_INT_TYPE,
   SLANG_LONG_TYPE
};

#define NUM_UINT_TYPES 4
static SLtype UInteger_Types[NUM_UINT_TYPES] =
{
   SLANG_UCHAR_TYPE,
   SLANG_USHORT_TYPE,
   SLANG_UINT_TYPE,
   SLANG_ULONG_TYPE
};

#define NUM_FLOAT_TYPES 2
static SLtype Float_Types[NUM_FLOAT_TYPES] =
{
   SLANG_FLOAT_TYPE,
   SLANG_DOUBLE_TYPE,
};

static double Double_NaN;

/*{{{ MPFR Interface Functions */

static int slMPFR_Error;
static int MPFR_Type_Id = -1;
static mp_rnd_t slMPFR_Rounding_Mode = GMP_RNDN;

typedef struct
{
   int numrefs;
   mpfr_t value;
}
MPFR_Type;

static MPFR_Type *MPFR_Type_NaN = NULL;

static int call_mpfr_init (mpfr_t x)
{
   if (setjmp (LongJmpEnv))
     return -1;

   mpfr_init (x);
   return 0;
}

static int call_mpfr_init_2_values (mpfr_t x, mpfr_t y)
{
   if (-1 == call_mpfr_init (x))
     return -1;

   if (-1 == call_mpfr_init (y))
     {
	mpfr_clear (x);
	return -1;
     }
   return 0;
}

#if 0				       /* currently unused */
static void call_mpfr_clear_2_values (mpfr_t a, mpfr_t b)
{
   mpfr_clear (a);
   mpfr_clear (b);
}

static int call_mpfr_init_4_values (mpfr_t a, mpfr_t b, mpfr_t c, mpfr_t d)
{
   if (-1 == call_mpfr_init_2_values (a, b))
     return -1;
   if (-1 == call_mpfr_init_2_values (c, d))
     {
	call_mpfr_clear_2_values (a, b);
	return -1;
     }
   return 0;
}

static void call_mpfr_clear_4_values (mpfr_t a, mpfr_t b, mpfr_t c, mpfr_t d)
{
   mpfr_clear (a);  mpfr_clear (b);
   mpfr_clear (c);  mpfr_clear (d);
}
#endif

static MPFR_Type *new_mpfr_type (void)
{
   MPFR_Type *mp;

   if (NULL == (mp = (MPFR_Type *)SLmalloc (sizeof (MPFR_Type))))
     return mp;

   if (-1 == call_mpfr_init (mp->value))
     {
	SLfree ((char *)mp);
	return NULL;
     }
   mp->numrefs = 1;
   return mp;
}

static void free_mpfr_type (MPFR_Type *mp)
{
   if (mp == NULL)
     return;
   if (mp->numrefs > 1)
     {
	mp->numrefs--;
	return;
     }
   mpfr_clear (mp->value);
   SLfree ((char *) mp);
}

static int push_mpfr_type (MPFR_Type *mp, int do_free)
{
   if (mp == NULL)
     return SLang_push_null ();

   if (do_free == 0)
     mp->numrefs++;

   if (0 == SLclass_push_ptr_obj (MPFR_Type_Id, (VOID_STAR)mp))
     return 0;

   if (do_free == 0)
     mp->numrefs--;
   else
     free_mpfr_type (mp);

   return -1;
}

/*{{{ Conversion of types to MPFR_Types */

static MPFR_Type *new_mpfr_type_from_string (char *str)
{
   MPFR_Type *mp;

   if (NULL == (mp = new_mpfr_type ()))
     return NULL;
   if (str == NULL)
     str = "nan";

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (mp);
	return NULL;
     }

   if (-1 == mpfr_set_str (mp->value, str, 10, slMPFR_Rounding_Mode))
     {
	SLang_verror (slMPFR_Error, "libmpfr: mpfr_set_str failed");
	free_mpfr_type (mp);
	return NULL;
     }

   return mp;
}

static MPFR_Type *new_mpfr_type_from_long (long x)
{
   MPFR_Type *mp;

   if (NULL == (mp = new_mpfr_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (mp);
	return NULL;
     }
   (void) mpfr_set_si (mp->value, x, slMPFR_Rounding_Mode);

   return mp;
}

static MPFR_Type *new_mpfr_type_from_ulong (unsigned long x)
{
   MPFR_Type *mp;

   if (NULL == (mp = new_mpfr_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (mp);
	return NULL;
     }

   (void) mpfr_set_ui (mp->value, x, slMPFR_Rounding_Mode);
   return mp;
}

static MPFR_Type *new_mpfr_type_from_double (double x)
{
   MPFR_Type *mp;

   if (NULL == (mp = new_mpfr_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (mp);
	return NULL;
     }

   (void) mpfr_set_d (mp->value, x, slMPFR_Rounding_Mode);
   return mp;
}

#if 0				       /* UNUSED */
static MPFR_Type *new_mpfr_type_from_mpfr (mpfr_t x)
{
   MPFR_Type *mp;

   if (NULL == (mp = new_mpfr_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (mp);
	return NULL;
     }

   (void) mpfr_set (mp->value, x, slMPFR_Rounding_Mode);
   return mp;
}
#endif

/*}}}*/

typedef long (*To_Long_Fun_Type)(VOID_STAR);
typedef unsigned long (*To_ULong_Fun_Type)(VOID_STAR);

#define TO_LONG_FUN(name, type) \
   static long name (VOID_STAR x) { return (long) *(type *) x; }
TO_LONG_FUN(char_to_long, char)
TO_LONG_FUN(short_to_long, short)
TO_LONG_FUN(int_to_long, int)
TO_LONG_FUN(long_to_long, long)

#define TO_ULONG_FUN(name, type) \
   static unsigned long name (VOID_STAR x) { return (unsigned long) *(type *) x; }
TO_ULONG_FUN(uchar_to_ulong, unsigned char)
TO_ULONG_FUN(ushort_to_ulong, unsigned short)
TO_ULONG_FUN(uint_to_ulong, unsigned int)
TO_ULONG_FUN(ulong_to_ulong, unsigned long)

static To_Long_Fun_Type get_to_long_fun (SLtype type, unsigned int *sizep)
{
   switch (type)
     {
      case SLANG_CHAR_TYPE: *sizep = sizeof(char); return char_to_long;
      case SLANG_SHORT_TYPE: *sizep = sizeof(short); return short_to_long;
      case SLANG_INT_TYPE: *sizep = sizeof(int); return int_to_long;
      case SLANG_LONG_TYPE: *sizep = sizeof(long); return long_to_long;
      default:
	SLang_verror (SL_Internal_Error, "unsupported type passed to get_to_long_fun");
	return NULL;
     }
}

static To_ULong_Fun_Type get_to_ulong_fun (SLtype type, unsigned int *sizep)
{
   switch (type)
     {
      case SLANG_UCHAR_TYPE: *sizep = sizeof(char); return uchar_to_ulong;
      case SLANG_USHORT_TYPE: *sizep = sizeof(unsigned short); return ushort_to_ulong;
      case SLANG_UINT_TYPE: *sizep = sizeof(unsigned int); return uint_to_ulong;
      case SLANG_ULONG_TYPE: *sizep = sizeof(unsigned long); return ulong_to_ulong;
      default:
	SLang_verror (SL_Internal_Error, "unsupported type passed to get_to_ulong_fun");
	return NULL;
     }
}


static int pop_mpfr_type (MPFR_Type **mpp)
{
#if 0
   /* This is handled by the typecast function */
   switch (SLang_peek_at_stack ())
     {
      case SLANG_STRING_TYPE:
      case SLANG_BSTRING_TYPE:
	  {
	     MPFR_Type *mp;
	     char *str;
	     if (-1 == SLang_pop_slstring (&str))
	       return -1;

	     if (NULL == (mp = new_mpfr_type_from_string (str)))
	       {
		  SLang_free_slstring (str);
		  return -1;
	       }
	     *mpp = mp;
	  }
	return 0;
     }
#endif
   return SLclass_pop_ptr_obj (MPFR_Type_Id, (VOID_STAR *)mpp);
}

static void mpfr_new_value (void)
{
   MPFR_Type *mp;
   if (SLang_peek_at_stack () != SLANG_ARRAY_TYPE)
     {
	if (0 == pop_mpfr_type (&mp))
	  (void) push_mpfr_type (mp, 1);
	return;
     }

   (void) SLclass_typecast (MPFR_Type_Id, 1, 1);
}

static unsigned long intrin_mpfr_get_def_prec (void)
{
   return mpfr_get_default_prec ();
}

static void intrin_mpfr_set_def_prec (unsigned long *pp)
{
   mpfr_prec_t p = (mpfr_prec_t) *pp;
   if (p < MPFR_PREC_MIN)
     p = 53;

   /* The docs say to stay away from the max precision */
   if (p >= MPFR_PREC_MAX - 32)
     p = MPFR_PREC_MAX-32;

   mpfr_set_default_prec (p);
}

static int do_intrin_mpfr_func0 (int (*func)(mpfr_t, mpfr_rnd_t))
{
   MPFR_Type *fr;

   if (NULL == (fr = new_mpfr_type ()))
     return -1;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (fr);
	return -1;
     }
   (void) (*func)(fr->value, slMPFR_Rounding_Mode);

   return push_mpfr_type (fr, 1);
}

static void intrin_mpfr_const_pi (void)
{
   (void) do_intrin_mpfr_func0 (mpfr_const_pi);
}

static void intrin_mpfr_const_log2 (void)
{
   (void) do_intrin_mpfr_func0 (mpfr_const_log2);
}

static unsigned long intrin_mpfr_get_prec (void)
{
   MPFR_Type *mp;
   mpfr_prec_t p;

   if (-1 == pop_mpfr_type (&mp))
     return 0;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (mp);
	return 0;
     }
   p = mpfr_get_prec (mp->value);
   free_mpfr_type (mp);
   return (unsigned long) p;
}

/*{{{ MPFR Class Methods */

static char *convert_mpfr_to_string (mpfr_t x) /*{{{*/
{
   char ebuf[256];		       /* bug enough to format a long integer */
   mp_exp_t expon;
   char *s;
   int shift_decimal;
   unsigned int len;
   char *xstr;

   if (mpfr_nan_p (x))
     return SLmake_string ("nan");
   if (mpfr_inf_p (x))
     {
	if (mpfr_sgn(x) >= 0)
	  return SLmake_string ("inf");
	else
	  return SLmake_string ("-inf");
     }
   if (mpfr_zero_p(x))
     {
	if (mpfr_sgn(x) >= 0)
	  return SLmake_string("0.0");
	else
	  return SLmake_string("-0.0");
     }

   s = mpfr_get_str (NULL, &expon, 10, 0, x, GMP_RNDD);
   if (s == NULL)
     return NULL;
   /* The generated string is a fraction, with an implicit radix point
    * immediately to the left of the first digit. For example, the
    * number −3.1416 would be returned as "−31416" in the string and
    * 1 written at expptr.
    *
    * In other words, s = abc...d ==> x = 0.abc...d*10^expon
    * Write this as: a.bc...dE{expon-1}.
    */
   shift_decimal = (expon && (expon - 1 < expon));
   if (shift_decimal)
     expon--;

   if (-1 == SLsnprintf (ebuf, sizeof(ebuf), "%ld", expon))
     {
	mpfr_free_str (s);
	return NULL;
     }
   /* need to allocate space for "0." + s + 'e' + expon + \0 */
   len = 2 + strlen (s) + 1 + strlen (ebuf) + 1;
   if (NULL == (xstr = SLmalloc(len)))
     {
	mpfr_free_str (s);
	return NULL;
     }
   if (shift_decimal == 0)
     {
	if (*s == '-')
	  (void) sprintf (xstr, "-0.%s", s+1);
	else
	  (void) sprintf (xstr, "0.%s", s);
     }
   else
     {
	if (*s == '-')
	  {
	     if (s[2] == 0)
	       (void) sprintf (xstr, "-%c", s[1]);
	     else
	       (void) sprintf (xstr, "-%c.%s", s[1], s+2);
	  }
	else
	  {
	     if (s[1] == 0)
	       (void) sprintf (xstr, "%c", s[0]);
	     else
	       (void) sprintf (xstr, "%c.%s", s[0], s+1);
	  }
     }
   mpfr_free_str (s);

   s = xstr + strlen(xstr);
   xstr++;
   while (s > xstr)
     {
	s--;
	if ((*s != '0') || (s[-1] == '.'))
	  break;
	*s = 0;
     }
   xstr--;

   if (expon != 0)
     sprintf (xstr+strlen(xstr), "e%s", ebuf);

   return xstr;
}

/*}}}*/

static void mpfr_type_destroy (SLtype type, VOID_STAR v)
{
   MPFR_Type *mp;

   (void) type;
   mp = *(MPFR_Type **) v;
   free_mpfr_type (mp);
}

static int mpfr_type_push (SLtype type, VOID_STAR v)
{
   (void) type;
   return push_mpfr_type (*(MPFR_Type **)v, 0);
}

static char *mpfr_type_string (SLtype type, VOID_STAR v)
{
   MPFR_Type *mp;

   (void) type;
   mp = *(MPFR_Type **)v;
   return convert_mpfr_to_string (mp->value);
}

static void free_mpfr_array (MPFR_Type **mpp, unsigned int n)
{
   unsigned int i;
   for (i = 0; i < n; i++)
     {
	free_mpfr_type (mpp[i]);
	mpp[i] = NULL;
     }
}

/*{{{ MPFR Binary Functions */

static int mpfr_binary_result (int op, SLtype a, SLtype b, SLtype *c)
{
   (void) a; (void) b;

   switch (op)
     {
      default:
	return 0;

      case SLANG_POW:
      case SLANG_PLUS:
      case SLANG_MINUS:
      case SLANG_TIMES:
      case SLANG_DIVIDE:
      case SLANG_MOD:
	*c = MPFR_Type_Id;
	break;

      case SLANG_EQ:
      case SLANG_NE:
      case SLANG_GT:
      case SLANG_GE:
      case SLANG_LT:
      case SLANG_LE:
	*c = SLANG_CHAR_TYPE;
	break;
     }
   return 1;
}

/*{{{ mpfr + mpfr etc functions  */

static MPFR_Type *call_mpfr_add (MPFR_Type *a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_add (c->value, a->value, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_sub (MPFR_Type *a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_sub (c->value, a->value, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_mul (MPFR_Type *a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_mul (c->value, a->value, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_pow (MPFR_Type *a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_pow (c->value, a->value, b->value, slMPFR_Rounding_Mode);
   return c;
}

#if MPFR_VERSION < MPFR_VERSION_NUM(2,4,0)
static int mpfr_fmod (mpfr_ptr r, mpfr_srcptr x, mpfr_srcptr y, mpfr_rnd_t rnd)
{
   /* r = x-y*n; n = ... -1, 0, 1, 2, ..
    * Want:
    *   If x > 0, 0 <= r < abs(y)
    *   If x < 0: r = -(abs(x) mod y)
    */
   mpfr_t xx;
   mpfr_t yy;
   int sx = mpfr_sgn(x);
   int sy = mpfr_sgn(y);

   if (sx < 0)
     {
	mpfr_init (xx);
	mpfr_abs (xx, x, rnd);
	if (sy < 0)
	  {
	     mpfr_init (yy);
	     mpfr_abs (yy, y, rnd);
	     mpfr_remainder (r, xx, yy, rnd);
	     if (mpfr_sgn(r) < 0)
	       mpfr_add (r, r, yy, GMP_RNDZ);

	     mpfr_setsign (r, r, -1, GMP_RNDZ);
	     mpfr_clear (yy);
	     mpfr_clear (xx);
	     return 0;
	  }

	mpfr_remainder (r, xx, y, rnd);
	if (mpfr_sgn(r) < 0)
	  mpfr_add (r, r, y, GMP_RNDZ);

	mpfr_setsign (r, r, -1, GMP_RNDZ);
	mpfr_clear (xx);
	return 0;
     }

   if (sy < 0)
     {
	mpfr_init (yy);
	mpfr_abs (yy, y, rnd);
	mpfr_remainder (r, x, yy, rnd);
	if (mpfr_sgn(r) < 0)
	  mpfr_add (r, r, yy, GMP_RNDZ);
	mpfr_clear (yy);
	return 0;
     }

   mpfr_remainder (r, x, y, rnd);
   if (mpfr_sgn(r) < 0)
     mpfr_add (r, r, y, GMP_RNDZ);

   return 0;
}
#endif


static MPFR_Type *call_mpfr_fmod (MPFR_Type *a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }

   mpfr_fmod (c->value, a->value, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_div (MPFR_Type *a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (b == NULL) b = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_div (c->value, a->value, b->value, slMPFR_Rounding_Mode);
   return c;
}

static int mpfr_mpfr_binary (int op, /*{{{*/
			     SLtype a_type, VOID_STAR ap, unsigned int na,
			     SLtype b_type, VOID_STAR bp, unsigned int nb,
			     VOID_STAR cp)
{
   MPFR_Type **a, **b, **c;
   unsigned int n, n_max;
   unsigned int da, db;
   char *ic;

   (void) a_type;
   (void) b_type;

   a = (MPFR_Type **) ap;
   b = (MPFR_Type **) bp;
   c = (MPFR_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_add (*a, *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_sub (*a, *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_mul (*a, *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_div (*a, *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a, *bb = *b;
	     if ((aa != NULL) && (bb != NULL))
	       ic[n] = mpfr_equal_p (aa->value, bb->value);
	     else
	       ic[n] = 0;
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a, *bb = *b;
	     if ((aa != NULL) && (bb != NULL))
	       ic[n] = !mpfr_equal_p (aa->value, bb->value);
	     else
	       ic[n] = 1;
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a, *bb = *b;
	     if ((aa != NULL) && (bb != NULL))
	       ic[n] = mpfr_greater_p (aa->value, bb->value);
	     else
	       ic[n] = 0;
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a, *bb = *b;
	     if ((aa != NULL) && (bb != NULL))
	       ic[n] = mpfr_greaterequal_p (aa->value, bb->value);
	     else
	       ic[n] = 0;
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a, *bb = *b;
	     if ((aa != NULL) && (bb != NULL))
	       ic[n] = mpfr_less_p (aa->value, bb->value);
	     else
	       ic[n] = 0;
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a, *bb = *b;
	     if ((aa != NULL) && (bb != NULL))
	       ic[n] = mpfr_lessequal_p (aa->value, bb->value);
	     else
	       ic[n] = 0;
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_pow (*a, *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MOD:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_fmod (*a, *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

	return 0;
     }

   return 1;
}

/*}}}*/

/*}}}*/

/*{{{ mpfr + double, double + mpfr, etc functions */

#if MPFR_VERSION < MPFR_VERSION_NUM(2,4,0)
/* The mpfr_add_d type functions were added to 2.4.0 */ /*{{{*/
static int mpfr_add_d (mpfr_ptr c, mpfr_srcptr a, double b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_add (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_sub_d (mpfr_ptr c, mpfr_srcptr a, double b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_sub (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_mul_d (mpfr_ptr c, mpfr_srcptr a, double b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_mul (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_div_d (mpfr_ptr c, mpfr_srcptr a, double b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_div (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_d_sub (mpfr_ptr c, double a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_d (aa, a, r);
   status = mpfr_sub (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}
static int mpfr_d_div (mpfr_ptr c, double a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_d (aa, a, r);
   status = mpfr_div (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}

/*}}}*/
#endif

static int mpfr_pow_d (mpfr_ptr c, mpfr_srcptr a, double b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_pow (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}

static int mpfr_d_pow (mpfr_ptr c, double a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_d (aa, a, r);
   status = mpfr_pow (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}

static int mpfr_fmod_d (mpfr_ptr c, mpfr_srcptr a, double b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_fmod (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}

static int mpfr_d_fmod (mpfr_ptr c, double a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_d (aa, a, r);
   status = mpfr_fmod (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}

static MPFR_Type *call_mpfr_add_d (MPFR_Type *a, double b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_add_d (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_sub_d (MPFR_Type *a, double b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_sub_d (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_mul_d (MPFR_Type *a, double b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_mul_d (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_div_d (MPFR_Type *a, double b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_div_d (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_d_sub (double a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_d_sub (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_d_div (double a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_d_div (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_pow_d (MPFR_Type *a, double b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_pow_d (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_d_pow (double a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_d_pow (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_fmod_d (MPFR_Type *a, double b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_fmod_d (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_d_fmod (double a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_d_fmod (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

/*}}}*/

/*{{{ mpfr + long, long + mpfr, etc functions */

#if MPFR_VERSION < MPFR_VERSION_NUM(2,4,0)
/* The mpfr_add_d type functions were added to 2.4.0 */ /*{{{*/
static int mpfr_add_d (mpfr_ptr c, mpfr_srcptr a, long b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_add (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_sub_d (mpfr_ptr c, mpfr_srcptr a, long b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_sub (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_mul_d (mpfr_ptr c, mpfr_srcptr a, long b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_mul (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_div_d (mpfr_ptr c, mpfr_srcptr a, long b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_div (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_d_sub (mpfr_ptr c, long a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_d (aa, a, r);
   status = mpfr_sub (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}
static int mpfr_d_div (mpfr_ptr c, long a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_d (aa, a, r);
   status = mpfr_div (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}

/*}}}*/
#endif

static int mpfr_si_pow (mpfr_ptr c, long a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_si (aa, a, r);
   status = mpfr_pow (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}

static int mpfr_fmod_si (mpfr_ptr c, mpfr_srcptr a, long b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_si (bb, b, r);
   status = mpfr_fmod (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}

static int mpfr_si_fmod (mpfr_ptr c, long a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_si (aa, a, r);
   status = mpfr_fmod (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}

static MPFR_Type *call_mpfr_add_si (MPFR_Type *a, long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_add_si (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_sub_si (MPFR_Type *a, long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_sub_si (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_mul_si (MPFR_Type *a, long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_mul_si (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_div_si (MPFR_Type *a, long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_div_si (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_si_sub (long a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_si_sub (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_si_div (long a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_si_div (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_pow_si (MPFR_Type *a, long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_pow_si (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_si_pow (long a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_si_pow (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_fmod_si (MPFR_Type *a, long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_fmod_si (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_si_fmod (long a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_si_fmod (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

/*}}}*/

/*{{{ mpfr + unsigned long, unsigned long + mpfr, etc functions */

#if MPFR_VERSION < MPFR_VERSION_NUM(2,4,0)
/* The mpfr_add_d type functions were added to 2.4.0 */ /*{{{*/
static int mpfr_add_d (mpfr_ptr c, mpfr_srcptr a, long b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_add (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_sub_d (mpfr_ptr c, mpfr_srcptr a, long b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_sub (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_mul_d (mpfr_ptr c, mpfr_srcptr a, long b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_mul (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_div_d (mpfr_ptr c, mpfr_srcptr a, long b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_d (bb, b, r);
   status = mpfr_div (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}
static int mpfr_d_sub (mpfr_ptr c, long a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_d (aa, a, r);
   status = mpfr_sub (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}
static int mpfr_d_div (mpfr_ptr c, long a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_d (aa, a, r);
   status = mpfr_div (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}

/*}}}*/
#endif

static int mpfr_fmod_ui (mpfr_ptr c, mpfr_srcptr a, unsigned long b, mpfr_rnd_t r)
{
   int status;
   mpfr_t bb;
   mpfr_init (bb);
   mpfr_set_ui (bb, b, r);
   status = mpfr_fmod (c, a, bb, r);
   mpfr_clear (bb);
   return status;
}

static int mpfr_ui_fmod (mpfr_ptr c, unsigned long a, mpfr_srcptr b, mpfr_rnd_t r)
{
   int status;
   mpfr_t aa;
   mpfr_init (aa);
   mpfr_set_ui (aa, a, r);
   status = mpfr_fmod (c, aa, b, r);
   mpfr_clear (aa);
   return status;
}

static MPFR_Type *call_mpfr_add_ui (MPFR_Type *a, unsigned long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_add_ui (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_sub_ui (MPFR_Type *a, unsigned long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_sub_ui (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_mul_ui (MPFR_Type *a, unsigned long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_mul_ui (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_div_ui (MPFR_Type *a, unsigned long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_div_ui (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_ui_sub (unsigned long a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_ui_sub (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_ui_div (unsigned long a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_ui_div (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_pow_ui (MPFR_Type *a, unsigned long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_pow_ui (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_ui_pow (unsigned long a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_ui_pow (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_fmod_ui (MPFR_Type *a, unsigned long b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_fmod_ui (c->value, a->value, b, slMPFR_Rounding_Mode);
   return c;
}

static MPFR_Type *call_mpfr_ui_fmod (unsigned long a, MPFR_Type *b)
{
   MPFR_Type *c = new_mpfr_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (c);
	return NULL;
     }
   mpfr_ui_fmod (c->value, a, b->value, slMPFR_Rounding_Mode);
   return c;
}

/*}}}*/

static int double_mpfr_binary (int op, /*{{{*/
			       SLtype a_type, VOID_STAR ap, unsigned int na,
			       SLtype b_type, VOID_STAR bp, unsigned int nb,
			       VOID_STAR cp)
{
   char *a;
   MPFR_Type **b, **c;
   char *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   SLang_To_Double_Fun_Type to_double;
   unsigned int sizeof_a;

   if (NULL == (to_double = SLarith_get_to_double_fun (a_type, &sizeof_a)))
     return 0;

   (void) b_type;

   a = (char *) ap;
   b = (MPFR_Type **) bp;
   c = (MPFR_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = sizeof_a;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_add_d (*b, to_double((VOID_STAR)a))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_d_sub (to_double((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_mul_d (*b, to_double((VOID_STAR)a))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_d_div (to_double((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) a);
	     if ((*b == NULL) || isnan(x) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_d ((*b)->value, x) == 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) a);
	     if ((*b == NULL) || isnan(x) || mpfr_nan_p((*b)->value))
	       ic[n] = 1;
	     else
	       ic[n] = (mpfr_cmp_d ((*b)->value, x) != 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) a);
	     if ((*b == NULL) || isnan(x) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_d ((*b)->value, x) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) a);
	     if ((*b == NULL) || isnan(x) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_d ((*b)->value, x) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) a);
	     if ((*b == NULL) || isnan(x) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_d ((*b)->value, x) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) a);
	     if ((*b == NULL) || isnan(x) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_d ((*b)->value, x) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_d_pow (to_double((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MOD:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_d_fmod (to_double((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;
     }

   return 1;
}

/*}}}*/

static int long_mpfr_binary (int op, /*{{{*/
			     SLtype a_type, VOID_STAR ap, unsigned int na,
			     SLtype b_type, VOID_STAR bp, unsigned int nb,
			     VOID_STAR cp)
{
   char *a;
   MPFR_Type **b, **c;
   char *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   long (*to_long)(VOID_STAR);
   unsigned int sizeof_a;

   if (NULL == (to_long = get_to_long_fun (a_type, &sizeof_a)))
     return 0;

   (void) b_type;

   a = (char *) ap;
   b = (MPFR_Type **) bp;
   c = (MPFR_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = sizeof_a;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_add_si (*b, to_long((VOID_STAR)a))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_si_sub (to_long((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_mul_si (*b, to_long((VOID_STAR)a))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_si_div (to_long((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_si ((*b)->value, x) == 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 1;
	     else
	       ic[n] = (mpfr_cmp_si ((*b)->value, x) != 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_si ((*b)->value, x) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_si ((*b)->value, x) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_si ((*b)->value, x) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_si ((*b)->value, x) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_si_pow (to_long((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MOD:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_si_fmod (to_long((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;
     }

   return 1;
}

/*}}}*/

static int ulong_mpfr_binary (int op, /*{{{*/
			     SLtype a_type, VOID_STAR ap, unsigned int na,
			     SLtype b_type, VOID_STAR bp, unsigned int nb,
			     VOID_STAR cp)
{
   char *a;
   MPFR_Type **b, **c;
   char *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   unsigned long (*to_ulong)(VOID_STAR);
   unsigned int sizeof_a;

   if (NULL == (to_ulong = get_to_ulong_fun (a_type, &sizeof_a)))
     return 0;

   (void) b_type;

   a = (char *) ap;
   b = (MPFR_Type **) bp;
   c = (MPFR_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = sizeof_a;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_add_ui (*b, to_ulong((VOID_STAR)a))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_ui_sub (to_ulong((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_mul_ui (*b, to_ulong((VOID_STAR)a))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_ui_div (to_ulong((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_ui ((*b)->value, x) == 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 1;
	     else
	       ic[n] = (mpfr_cmp_ui ((*b)->value, x) != 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_ui ((*b)->value, x) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_ui ((*b)->value, x) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_ui ((*b)->value, x) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) a);
	     if ((*b == NULL) || mpfr_nan_p((*b)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_ui ((*b)->value, x) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_ui_pow (to_ulong((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MOD:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_ui_fmod (to_ulong((VOID_STAR)a), *b)))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;
     }

   return 1;
}

/*}}}*/

static int mpfr_double_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   MPFR_Type **a, **c;
   char *b, *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   SLang_To_Double_Fun_Type to_double;
   unsigned int sizeof_b;

   if (NULL == (to_double = SLarith_get_to_double_fun (b_type, &sizeof_b)))
     return 0;

   (void) a_type;

   a = (MPFR_Type **) ap;
   b = (char *) bp;
   c = (MPFR_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = sizeof_b;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_add_d (*a, to_double((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_sub_d (*a, to_double((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_mul_d (*a, to_double((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_div_d (*a, to_double((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_d ((*a)->value, x) == 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 1;
	     else
	       ic[n] = (mpfr_cmp_d ((*a)->value, x) != 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_d ((*a)->value, x) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_d ((*a)->value, x) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_d ((*a)->value, x) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double x = to_double ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_d ((*a)->value, x) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_pow_d (*a, to_double((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MOD:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_fmod_d (*a, to_double((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;
     }

   return 1;
}

/*}}}*/

static int mpfr_long_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   MPFR_Type **a, **c;
   char *b, *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   To_Long_Fun_Type to_long;
   unsigned int sizeof_b;

   if (NULL == (to_long = get_to_long_fun (b_type, &sizeof_b)))
     return 0;

   (void) a_type;

   a = (MPFR_Type **) ap;
   b = (char *) bp;
   c = (MPFR_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = sizeof_b;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_add_si (*a, to_long((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_sub_si (*a, to_long((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_mul_si (*a, to_long((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_div_si (*a, to_long((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) b);
	     if ((*a == NULL) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_si ((*a)->value, x) == 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) b);
	     if ((*a == NULL) || mpfr_nan_p((*a)->value))
	       ic[n] = 1;
	     else
	       ic[n] = (mpfr_cmp_si ((*a)->value, x) != 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) b);
	     if ((*a == NULL) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_si ((*a)->value, x) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) b);
	     if ((*a == NULL) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_si ((*a)->value, x) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) b);
	     if ((*a == NULL) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_si ((*a)->value, x) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long x = to_long ((VOID_STAR) b);
	     if ((*a == NULL) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_si ((*a)->value, x) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_pow_si (*a, to_long((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MOD:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_fmod_si (*a, to_long((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;
     }

   return 1;
}

/*}}}*/

static int mpfr_ulong_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   MPFR_Type **a, **c;
   char *b, *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   To_ULong_Fun_Type to_ulong;
   unsigned int sizeof_b;

   if (NULL == (to_ulong = get_to_ulong_fun (b_type, &sizeof_b)))
     return 0;

   (void) a_type;

   a = (MPFR_Type **) ap;
   b = (char *) bp;
   c = (MPFR_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = sizeof_b;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_add_ui (*a, to_ulong((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_sub_ui (*a, to_ulong((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_mul_ui (*a, to_ulong((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_div_ui (*a, to_ulong((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_ui ((*a)->value, x) == 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 1;
	     else
	       ic[n] = (mpfr_cmp_ui ((*a)->value, x) != 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_ui ((*a)->value, x) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_ui ((*a)->value, x) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_ui ((*a)->value, x) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long x = to_ulong ((VOID_STAR) b);
	     if ((*a == NULL) || isnan(x) || mpfr_nan_p((*a)->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfr_cmp_ui ((*a)->value, x) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_pow_ui (*a, to_ulong((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MOD:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfr_fmod_ui (*a, to_ulong((VOID_STAR)b))))
	       {
		  free_mpfr_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;
     }

   return 1;
}

/*}}}*/

/*}}}*/

/*{{{ MPFR Unary Functions */

static int mpfr_unary_result (int op, SLtype a, SLtype *b)
{
   (void) a;

   switch (op)
     {
      case SLANG_PLUSPLUS:
      case SLANG_MINUSMINUS:
      case SLANG_SIGN:
      default:
	return 0;

      case SLANG_CHS:
      case SLANG_MUL2:
      case SLANG_SQR:
      case SLANG_ABS:
	*b = MPFR_Type_Id;
	break;

      case SLANG_ISPOS:
      case SLANG_ISNEG:
      case SLANG_ISNONNEG:
	*b = SLANG_CHAR_TYPE;
	break;
     }
   return 1;
}

static MPFR_Type *call_mpfr_neg (MPFR_Type *a)
{
   MPFR_Type *b = new_mpfr_type();
   if (b == NULL) return NULL;

   if (a == NULL) a = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (b);
	return NULL;
     }
   (void) mpfr_neg (b->value, a->value, slMPFR_Rounding_Mode);
   return b;
}

static MPFR_Type *call_mpfr_sqr (MPFR_Type *a)
{
   MPFR_Type *b = new_mpfr_type();
   if (b == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (b);
	return NULL;
     }
   (void) mpfr_sqr (b->value, a->value, slMPFR_Rounding_Mode);
   return b;
}

static MPFR_Type *call_mpfr_abs (MPFR_Type *a)
{
   MPFR_Type *b = new_mpfr_type();
   if (b == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (b);
	return NULL;
     }
   (void) mpfr_abs (b->value, a->value, slMPFR_Rounding_Mode);
   return b;
}

static int mpfr_unary (int op,
		       SLtype a_type, VOID_STAR ap, unsigned int na,
		       VOID_STAR bp)
{
   unsigned int n;
   MPFR_Type **a, **b;
   char *c;
   MPFR_Type *two;

   (void) a_type;
   a = (MPFR_Type **) ap;
   b = (MPFR_Type **) bp;

   switch (op)
     {
      case SLANG_PLUSPLUS:
      case SLANG_MINUSMINUS:
      case SLANG_SIGN:
      default:
	return 0;

      case SLANG_CHS:
	for (n = 0; n < na; n++)
	  {
	     if (NULL == (b[n] = call_mpfr_neg (a[n])))
	       {
		  free_mpfr_array (b, n);
		  return -1;
	       }
	  }
	break;

      case SLANG_SQR:
	for (n = 0; n < na; n++)
	  {
	     if (NULL == (b[n] = call_mpfr_sqr (a[n])))
	       {
		  free_mpfr_array (b, n);
		  return -1;
	       }
	  }
	break;

      case SLANG_MUL2:
	if (NULL == (two = new_mpfr_type_from_double (2.0)))
	  return -1;

	for (n = 0; n < na; n++)
	  {
	     if (NULL == (b[n] = call_mpfr_mul (two, a[n])))
	       {
		  free_mpfr_type (two);
		  free_mpfr_array (b, n);
		  return -1;
	       }
	  }
	free_mpfr_type (two);
	break;

      case SLANG_ABS:
	for (n = 0; n < na; n++)
	  {
	     if (NULL == (b[n] = call_mpfr_abs (a[n])))
	       {
		  free_mpfr_array (b, n);
		  return -1;
	       }
	  }
	break;

      case SLANG_ISPOS:
	c = (char *) bp;
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < na; n++)
	  {
	     if (b[n] == NULL)
	       c[n] = 0;
	     else
	       c[n] = (mpfr_sgn(b[n]->value) > 0);
	  }
	break;

      case SLANG_ISNEG:
	c = (char *) bp;
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < na; n++)
	  {
	     if (b[n] == NULL)
	       c[n] = 0;
	     else
	       c[n] = (mpfr_sgn(b[n]->value) < 0);
	  }
	break;

      case SLANG_ISNONNEG:
	c = (char *) bp;
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < na; n++)
	  {
	     if (b[n] == NULL)
	       c[n] = 0;
	     else
	       c[n] = (mpfr_sgn(b[n]->value) >= 0);
	  }
	break;
     }
   return 1;
}

/*}}}*/

/*{{{ MPFR Math Unary Functions */

static MPFR_Type *call_mpfr_int_func2 (int (*func)(mpfr_ptr, mpfr_srcptr, mp_rnd_t), MPFR_Type *a)
{
   MPFR_Type *b = new_mpfr_type();
   if (b == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfr_type (b);
	return NULL;
     }
   (void) (*func) (b->value, a->value, slMPFR_Rounding_Mode);
   return b;
}

static int mpfr_math_unary_result (int op, SLtype a, SLtype *b) /*{{{*/
{
   (void) a;

   switch (op)
     {
      case SLMATH_SINH:
      case SLMATH_COSH:
      case SLMATH_TANH:
      case SLMATH_TAN:
      case SLMATH_ASIN:
      case SLMATH_ACOS:
      case SLMATH_ATAN:
      case SLMATH_EXP:
      case SLMATH_LOG:
      case SLMATH_LOG10:
      case SLMATH_SQRT:
      case SLMATH_SIN:
      case SLMATH_COS:
      case SLMATH_ASINH:
      case SLMATH_ATANH:
      case SLMATH_ACOSH:
      case SLMATH_EXPM1:
      case SLMATH_LOG1P:
	*b = MPFR_Type_Id;
	break;

      case SLMATH_ISINF:
      case SLMATH_ISNAN:
	*b = SLANG_CHAR_TYPE;
	break;

      case SLMATH_FLOOR:
      case SLMATH_CEIL:
      case SLMATH_ROUND:
	*b = MPFR_Type_Id;
	break;

      default:
	return 0;
     }

   return 1;
}

/*}}}*/

static int do_mpfr_floor (mpfr_ptr b, mpfr_srcptr a, mp_rnd_t r)
{
   (void) r;
   return mpfr_floor (b, a);
}

static int do_mpfr_ceil (mpfr_ptr b, mpfr_srcptr a, mp_rnd_t r)
{
   (void) r;
   return mpfr_ceil (b, a);
}

static int do_mpfr_round (mpfr_ptr b, mpfr_srcptr a, mp_rnd_t r)
{
   (void) r;
   return mpfr_round (b, a);
}

static int mpfr_math_unary (int op, /*{{{*/
			    SLtype a_type, VOID_STAR ap, unsigned int na,
			    VOID_STAR bp)
{
   unsigned int i;
   MPFR_Type **a, **b;
   int (*func)(mpfr_ptr, mpfr_srcptr, mp_rnd_t);

   (void) a_type;
   a = (MPFR_Type **) ap;
   b = (MPFR_Type **) bp;

   switch (op)
     {
      case SLMATH_SINH: func = mpfr_sinh; break;
      case SLMATH_COSH: func = mpfr_cosh; break;
      case SLMATH_TANH: func = mpfr_tanh; break;
      case SLMATH_TAN: func = mpfr_tan; break;
      case SLMATH_ASIN: func = mpfr_asin; break;
      case SLMATH_ACOS: func = mpfr_acos; break;
      case SLMATH_ATAN: func = mpfr_atan; break;
      case SLMATH_EXP: func = mpfr_exp; break;
      case SLMATH_LOG: func = mpfr_log; break;
      case SLMATH_LOG10: func = mpfr_log10; break;
      case SLMATH_SQRT: func = mpfr_sqrt; break;
      case SLMATH_SIN: func = mpfr_sin; break;
      case SLMATH_COS: func = mpfr_cos; break;
      case SLMATH_ASINH: func = mpfr_asinh; break;
      case SLMATH_ATANH: func = mpfr_atanh; break;
      case SLMATH_ACOSH: func = mpfr_acosh; break;
      case SLMATH_EXPM1: func = mpfr_expm1; break;
      case SLMATH_LOG1P: func = mpfr_log1p; break;

      case SLMATH_ISINF:
	  {
	     char *c = (char *)bp;
	     if (setjmp (LongJmpEnv))
	       return -1;
	     for (i = 0; i < na; i++)
	       {
		  if (a[i] == NULL)
		    c[i] = 0;
		  else
		    c[i] = (char) mpfr_inf_p (a[i]->value);
	       }
	  }
	return 1;

      case SLMATH_ISNAN:
	  {
	     char *c = (char *)bp;
	     if (setjmp (LongJmpEnv))
	       return -1;
	     for (i = 0; i < na; i++)
	       {
		  if (a[i] == NULL)
		    c[i] = 1;
		  else
		    c[i] = (char) mpfr_nan_p (a[i]->value);
	       }
	  }
	return 1;

      case SLMATH_FLOOR:
	func = do_mpfr_floor;
	break;
      case SLMATH_CEIL:
	func = do_mpfr_ceil;
	break;
      case SLMATH_ROUND:
	func = do_mpfr_round;
	break;

      default:
	return 0;
     }

   for (i = 0; i < na; i++)
     {
	if (NULL == (b[i] = call_mpfr_int_func2 (func, a[i])))
	  {
	     free_mpfr_array (b, i);
	     return -1;
	  }
     }

   return 1;
}

/*}}}*/

/*}}}*/

#define TYPECAST_BODY(type, func) \
   { \
      type *ptr = (type *) from; \
      for (n = 0; n < num; n++) \
	{ \
	   MPFR_Type *mp; \
	   type x = *ptr++; \
	   if (NULL == (mp = func (x))) \
	     { \
		free_mpfr_array (mpp, n); \
		return -1; \
	     } \
	   *mpp++ = mp; \
	} \
   }

static int mpfr_typecast (SLtype from_type, VOID_STAR from, unsigned int num, /*{{{*/
	       SLtype to_type, VOID_STAR to)
{
   unsigned int n;
   MPFR_Type **mpp;

   (void) to_type;

   mpp = (MPFR_Type **) to;
   memset ((char *)mpp, 0, num*sizeof(MPFR_Type *));

   switch (from_type)
     {
      case SLANG_UCHAR_TYPE:
	TYPECAST_BODY(unsigned char, new_mpfr_type_from_ulong) return 1;
      case SLANG_USHORT_TYPE:
	TYPECAST_BODY(unsigned short, new_mpfr_type_from_ulong) return 1;
      case SLANG_UINT_TYPE:
	TYPECAST_BODY(unsigned int, new_mpfr_type_from_ulong) return 1;
      case SLANG_ULONG_TYPE:
	TYPECAST_BODY(unsigned long, new_mpfr_type_from_ulong) return 1;

      case SLANG_CHAR_TYPE:
	TYPECAST_BODY(char, new_mpfr_type_from_long) return 1;
      case SLANG_SHORT_TYPE:
	TYPECAST_BODY(short, new_mpfr_type_from_long) return 1;
      case SLANG_INT_TYPE:
	TYPECAST_BODY(int, new_mpfr_type_from_long) return 1;
      case SLANG_LONG_TYPE:
	TYPECAST_BODY(long, new_mpfr_type_from_long) return 1;

      case SLANG_FLOAT_TYPE:
	TYPECAST_BODY(float, new_mpfr_type_from_double) return 1;
      case SLANG_DOUBLE_TYPE:
	TYPECAST_BODY(double, new_mpfr_type_from_double) return 1;

      case SLANG_STRING_TYPE:
	TYPECAST_BODY(char *, new_mpfr_type_from_string) return 1;

      case SLANG_BSTRING_TYPE:
	  {
	     SLang_BString_Type **ptr;

	     ptr = (SLang_BString_Type **) from;

	     for (n = 0; n < num; n++)
	       {
		  MPFR_Type *mp;
		  char *s;
		  unsigned int len;

		  s = (char *)SLbstring_get_pointer (*ptr, &len);
		  if (NULL == (mp = new_mpfr_type_from_string (s)))
		    {
		       free_mpfr_array (mpp, n);
		       return -1;
		    }
		  *mpp++ = mp;
		  ptr++;
	       }
	  }
	return 1;

      default:
	break;
     }
   return 0;
}
#undef TYPECAST_BODY

static int mpfr_to_double_typecast (SLtype from_type, VOID_STAR from, unsigned int num, /*{{{*/
				    SLtype to_type, VOID_STAR to)
{
   unsigned int i;
   MPFR_Type **mpp;
   double *dp;

   (void) from_type;
   (void) to_type;

   mpp = (MPFR_Type **) from;
   dp = (double *) to;

   if (setjmp(LongJmpEnv))
     return -1;
   for (i = 0; i < num; i++)
     {
	MPFR_Type *mp = mpp[i];
	if (mp == NULL)
	  dp[i] = Double_NaN;
	else
	  dp[i] = mpfr_get_d (mp->value, slMPFR_Rounding_Mode);
     }
   return 1;
}

/*}}}*/

/*}}}*/

static int init_mpfr_class (void) /*{{{*/
{
   SLang_Class_Type *cl;
   unsigned int i;

   if (MPFR_Type_Id != -1)
     return 0;

   if (NULL == (cl = SLclass_allocate_class ("MPFR_Type")))
     return -1;

   (void) SLclass_set_string_function (cl, mpfr_type_string);
   (void) SLclass_set_push_function (cl, mpfr_type_push);
   (void) SLclass_set_destroy_function (cl, mpfr_type_destroy);

   if (-1 == SLclass_register_class (cl, SLANG_VOID_TYPE, sizeof (VOID_STAR),
				     SLANG_CLASS_TYPE_PTR))
     return -1;

   MPFR_Type_Id = SLclass_get_class_id (cl);

   for (i = 0; i < NUM_INT_TYPES; i++)
     {
	SLtype t = Integer_Types[i];
	if (-1 == (SLclass_add_typecast(t, MPFR_Type_Id, mpfr_typecast, 1)))
	  return -1;
	if (-1 == SLclass_add_binary_op (t, MPFR_Type_Id, long_mpfr_binary, mpfr_binary_result))
	  return -1;
	if (-1 == SLclass_add_binary_op (MPFR_Type_Id, t, mpfr_long_binary, mpfr_binary_result))
	  return -1;
     }
   for (i = 0; i < NUM_UINT_TYPES; i++)
     {
	SLtype t = UInteger_Types[i];
	if (-1 == (SLclass_add_typecast(t, MPFR_Type_Id, mpfr_typecast, 1)))
	  return -1;
	if (-1 == SLclass_add_binary_op (t, MPFR_Type_Id, ulong_mpfr_binary, mpfr_binary_result))
	  return -1;
	if (-1 == SLclass_add_binary_op (MPFR_Type_Id, t, mpfr_ulong_binary, mpfr_binary_result))
	  return -1;
     }
   for (i = 0; i < NUM_FLOAT_TYPES; i++)
     {
	SLtype t = Float_Types[i];
	if (-1 == (SLclass_add_typecast(t, MPFR_Type_Id, mpfr_typecast, 1)))
	  return -1;
	if (-1 == SLclass_add_binary_op (t, MPFR_Type_Id, double_mpfr_binary, mpfr_binary_result))
	  return -1;
	if (-1 == SLclass_add_binary_op (MPFR_Type_Id, t, mpfr_double_binary, mpfr_binary_result))
	  return -1;
     }

   if (-1 == (SLclass_add_typecast(MPFR_Type_Id, SLANG_DOUBLE_TYPE, mpfr_to_double_typecast, 1)))
     return -1;

   if ((-1 == SLclass_add_typecast (SLANG_STRING_TYPE, MPFR_Type_Id, mpfr_typecast, 1))
       || (-1 == SLclass_add_typecast (SLANG_BSTRING_TYPE, MPFR_Type_Id, mpfr_typecast, 1)))
     return -1;

   if (-1 == SLclass_add_binary_op (MPFR_Type_Id, MPFR_Type_Id, mpfr_mpfr_binary, mpfr_binary_result))
     return -1;

   if (-1 == (SLclass_add_unary_op (MPFR_Type_Id, mpfr_unary, mpfr_unary_result)))
     return -1;

   if (-1 == SLclass_add_math_op (MPFR_Type_Id, mpfr_math_unary, mpfr_math_unary_result))
     return -1;

   slMPFR_Error = SLerr_new_exception (SL_RunTime_Error, "MPFR_Error", "MPFR Error");

   if (NULL == (MPFR_Type_NaN = new_mpfr_type ()))
     return -1;

   return 0;
}

/*}}}*/

/*}}}*/

/*}}}*/

/*{{{ MPFI Interface Functions */

static int MPFI_Type_Id = -1;
static int slMPFI_Error;
typedef struct
{
   int numrefs;
   mpfi_t value;
}
MPFI_Type;

static MPFI_Type *MPFI_Type_NaN = NULL;

static MPFI_Type *new_mpfi_type (void)
{
   MPFI_Type *mp;

   if (NULL == (mp = (MPFI_Type *)SLmalloc (sizeof (MPFI_Type))))
     return mp;

   if (setjmp (LongJmpEnv))
     {
	SLfree ((char *)mp);
	return NULL;
     }

   mpfi_init (mp->value);
   mp->numrefs = 1;
   return mp;
}

static void free_mpfi_type (MPFI_Type *mp)
{
   if (mp == NULL)
     return;
   if (mp->numrefs > 1)
     {
	mp->numrefs--;
	return;
     }
   mpfi_clear (mp->value);
   SLfree ((char *) mp);
}

static int push_mpfi_type (MPFI_Type *mp, int do_free)
{
   if (mp == NULL)
     return SLang_push_null ();

   if (do_free == 0)
     mp->numrefs++;

   if (0 == SLclass_push_ptr_obj (MPFI_Type_Id, (VOID_STAR)mp))
     return 0;

   if (do_free == 0)
     mp->numrefs--;
   else
     free_mpfi_type (mp);

   return -1;
}

static MPFI_Type *new_mpfi_type_from_string (char *str)
{
   MPFI_Type *mp;
   int status;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;
   if (str == NULL)
     str = "nan";

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }

   status = mpfi_set_str (mp->value, str, 10);
   if (status == 0)
     return mp;

   free_mpfi_type (mp);
   if (status == 1)
     SLang_verror (slMPFI_Error, "libmpfi: Error parsing input");
   else
     SLang_verror (slMPFI_Error, "libmpfi: internal overflow");

   return NULL;
}

static MPFI_Type *new_mpfi_type_from_long (long x)
{
   MPFI_Type *mp;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   (void) mpfi_set_si (mp->value, x);

   return mp;
}

static MPFI_Type *new_mpfi_type_from_ulong (unsigned long x)
{
   MPFI_Type *mp;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }

   (void) mpfi_set_ui (mp->value, x);
   return mp;
}

static MPFI_Type *new_mpfi_type_from_double (double x)
{
   MPFI_Type *mp;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }

   (void) mpfi_set_d (mp->value, x);
   return mp;
}

static MPFI_Type *new_mpfi_type_from_mpfr_type (MPFR_Type *fr)
{
   MPFI_Type *mp;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;
   if (fr == NULL) fr = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   (void) mpfi_set_fr (mp->value, fr->value);
   return mp;
}

static int pop_mpfi_type (MPFI_Type **mpp)
{
   switch (SLang_peek_at_stack ())
     {
      case SLANG_STRING_TYPE:
      case SLANG_BSTRING_TYPE:
	  {
	     char *str;
	     MPFI_Type *mp;

	     if (-1 == SLang_pop_slstring (&str))
	       return -1;

	     if (NULL == (mp = new_mpfi_type_from_string (str)))
	       {
		  SLang_free_slstring (str);
		  return -1;
	       }
	     *mpp = mp;
	  }
	return 0;

      default:
	break;
     }
   return SLclass_pop_ptr_obj (MPFI_Type_Id, (VOID_STAR *)mpp);
}

/* This routines will return arrays of the following types:
 *   long, double, mpfr, unsigned long.
 */
static int pop_mpfr_array_internal (SLang_Array_Type **atp)
{
   switch (SLang_peek_at_stack1 ())
     {
      default:
	if (-1 == SLclass_typecast (MPFR_Type_Id, 1, 1))
	  return -1;
	/* drop */
      case SLANG_INT_TYPE:
      case SLANG_UINT_TYPE:
      case SLANG_DOUBLE_TYPE:
      case SLANG_LONG_TYPE:
	return SLang_pop_array (atp, 0);

      case SLANG_UCHAR_TYPE:
      case SLANG_USHORT_TYPE:
      case SLANG_ULONG_TYPE:
	return SLang_pop_array_of_type (atp, SLANG_ULONG_TYPE);
     }
}

static MPFI_Type *new_mpfi_from_l_l (long la, long lb, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   (void) a; (void) b;
   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   (void) mpfi_interv_si (mp->value, la, lb);
   return mp;
}

static MPFI_Type *new_mpfi_from_l_u (long la, unsigned long ub, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   mpfr_set_si (a, la, slMPFR_Rounding_Mode);
   mpfr_set_ui (b, ub, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, a, b);
   return mp;
}

static MPFI_Type *new_mpfi_from_l_d (long la, double db, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   mpfr_set_si (a, la, slMPFR_Rounding_Mode);
   mpfr_set_d (b, db, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, a, b);
   return mp;
}

static MPFI_Type *new_mpfi_from_l_fr (long la, MPFR_Type *fr_b, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   (void) b;
   if (NULL == (mp = new_mpfi_type ()))
     return NULL;
   if (fr_b == NULL) fr_b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   mpfr_set_si (a, la, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, a, fr_b->value);
   return mp;
}

static MPFI_Type *new_mpfi_from_u_l (unsigned long ua, long lb, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   (void) mpfr_set_ui (a, ua, slMPFR_Rounding_Mode);
   (void) mpfr_set_si (b, lb, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, a, b);
   return mp;
}

static MPFI_Type *new_mpfi_from_u_u (unsigned long ua, unsigned long ub, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   (void) a; (void) b;
   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   (void) mpfi_interv_ui (mp->value, ua, ub);
   return mp;
}

static MPFI_Type *new_mpfi_from_u_d (unsigned long ua, double db, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   mpfr_set_ui (a, ua, slMPFR_Rounding_Mode);
   mpfr_set_d (b, db, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, a, b);
   return mp;
}

static MPFI_Type *new_mpfi_from_u_fr (unsigned long ua, MPFR_Type *fr_b, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   (void) b;
   if (fr_b == NULL) fr_b = MPFR_Type_NaN;
   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   mpfr_set_ui (a, ua, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, a, fr_b->value);
   return mp;
}

static MPFI_Type *new_mpfi_from_d_l (double da, long lb, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   (void) mpfr_set_d (a, da, slMPFR_Rounding_Mode);
   (void) mpfr_set_si (b, lb, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, a, b);
   return mp;
}

static MPFI_Type *new_mpfi_from_d_u (double da, unsigned long ub, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   (void) mpfr_set_d (a, da, slMPFR_Rounding_Mode);
   (void) mpfr_set_ui (b, ub, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, a, b);
   return mp;
}

static MPFI_Type *new_mpfi_from_d_d (double da, double db, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   (void) a; (void) b;
   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   (void) mpfi_interv_d (mp->value, da, db);
   return mp;
}

static MPFI_Type *new_mpfi_from_d_fr (double da, MPFR_Type *fr_b, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   (void) b;
   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (fr_b == NULL) fr_b = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   mpfr_set_d (a, da, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, a, fr_b->value);
   return mp;
}

static MPFI_Type *new_mpfi_from_fr_l (MPFR_Type *fr_a, long lb, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   (void) a;
   if (NULL == (mp = new_mpfi_type ()))
     return NULL;
   if (fr_a == NULL) fr_a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   (void) mpfr_set_si (b, lb, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, fr_a->value, b);
   return mp;
}

static MPFI_Type *new_mpfi_from_fr_u (MPFR_Type *fr_a, unsigned long ub, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   (void) a;
   if (NULL == (mp = new_mpfi_type ()))
     return NULL;
   if (fr_a == NULL) fr_a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   mpfr_set_ui (b, ub, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, fr_a->value, b);
   return mp;
}

static MPFI_Type *new_mpfi_from_fr_d (MPFR_Type *fr_a, double db, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   (void) a;

   if (NULL == (mp = new_mpfi_type ()))
     return NULL;
   if (fr_a == NULL) fr_a = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   mpfr_set_d (b, db, slMPFR_Rounding_Mode);
   (void) mpfi_interv_fr (mp->value, fr_a->value, b);
   return mp;
}

static MPFI_Type *new_mpfi_from_fr_fr (MPFR_Type *fr_a, MPFR_Type *fr_b, mpfr_t a, mpfr_t b)
{
   MPFI_Type *mp;

   (void) a; (void) b;
   if (NULL == (mp = new_mpfi_type ()))
     return NULL;

   if (fr_a == NULL) fr_a = MPFR_Type_NaN;
   if (fr_b == NULL) fr_b = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (mp);
	return NULL;
     }
   (void) mpfi_interv_fr (mp->value, fr_a->value, fr_b->value);
   return mp;
}

/* at_a, at_b must be one of the following types:
 *   int long, double, mpfr, unsigned int, unsigned long.
 */
static int set_mpfi_array_values_from_arrays (SLang_Array_Type *at_c,
					      SLang_Array_Type *at_a,
					      SLang_Array_Type *at_b)
{
   SLuindex_Type i, num;
   mpfr_t a, b;
   MPFI_Type **cp;
   MPFR_Type **fr_a, **fr_b;
   double *d_a, *d_b;
   unsigned long *ul_a, *ul_b;
   long *l_a, *l_b;
   int *i_a, *i_b;
   unsigned int *ui_a, *ui_b;
   int status = -1;

   call_mpfr_init_2_values (a, b);

   num = at_c->num_elements;
   cp = (MPFI_Type **) at_c->data;
#define CASE_BODY(func,xa,xb,a,b) \
   for (i = 0; i < num; i++) \
   { \
      if (NULL == (cp[i] = func (xa[i], xb[i], a, b))) \
	goto free_and_return; \
   } status = 0

   switch (at_a->data_type)
     {
      case SLANG_INT_TYPE:
	i_a = (int *)at_a->data;
	switch (at_b->data_type)
	  {
	   case SLANG_INT_TYPE:
	     i_b = (int *) at_b->data;
	     CASE_BODY(new_mpfi_from_l_l, i_a, i_b, a, b);
	     break;

	   case SLANG_LONG_TYPE:
	     l_b = (long *) at_b->data;
	     CASE_BODY(new_mpfi_from_l_l, i_a, l_b, a, b);
	     break;

	   case SLANG_UINT_TYPE:
	     ui_b = (unsigned int *) at_b->data;
	     CASE_BODY(new_mpfi_from_l_u, i_a, ui_b, a, b);
	     break;

	   case SLANG_ULONG_TYPE:
	     ul_b = (unsigned long *) at_b->data;
	     CASE_BODY(new_mpfi_from_l_u, i_a, ul_b, a, b);
	     break;

	   case SLANG_DOUBLE_TYPE:
	     d_b = (double *) at_b->data;
	     CASE_BODY(new_mpfi_from_l_d, i_a, d_b, a, b);
	     break;

	   default:		       /* MPFR_Type */
	     fr_b = (MPFR_Type **) at_b->data;
	     CASE_BODY(new_mpfi_from_l_fr, i_a, fr_b, a, b);
	     break;
	  }
	break;

      case SLANG_LONG_TYPE:
	l_a = (long *)at_a->data;
	switch (at_b->data_type)
	  {
	   case SLANG_INT_TYPE:
	     i_b = (int *) at_b->data;
	     CASE_BODY(new_mpfi_from_l_l, l_a, i_b, a, b);
	     break;

	   case SLANG_LONG_TYPE:
	     l_b = (long *) at_b->data;
	     CASE_BODY(new_mpfi_from_l_l, l_a, l_b, a, b);
	     break;

	   case SLANG_UINT_TYPE:
	     ui_b = (unsigned int *) at_b->data;
	     CASE_BODY(new_mpfi_from_l_u, l_a, ui_b, a, b);
	     break;

	   case SLANG_ULONG_TYPE:
	     ul_b = (unsigned long *) at_b->data;
	     CASE_BODY(new_mpfi_from_l_u, l_a, ul_b, a, b);
	     break;

	   case SLANG_DOUBLE_TYPE:
	     d_b = (double *) at_b->data;
	     CASE_BODY(new_mpfi_from_l_d, l_a, d_b, a, b);
	     break;

	   default:		       /* MPFR_Type */
	     fr_b = (MPFR_Type **) at_b->data;
	     CASE_BODY(new_mpfi_from_l_fr, l_a, fr_b, a, b);
	     break;
	  }
	break;

      case SLANG_UINT_TYPE:
	ui_a = (unsigned int *)at_a->data;
	switch (at_b->data_type)
	  {
	   case SLANG_INT_TYPE:
	     i_b = (int *) at_b->data;
	     CASE_BODY(new_mpfi_from_u_l, ui_a, i_b, a, b);
	     break;

	   case SLANG_LONG_TYPE:
	     l_b = (long *) at_b->data;
	     CASE_BODY(new_mpfi_from_u_l, ui_a, l_b, a, b);
	     break;

	   case SLANG_UINT_TYPE:
	     ui_b = (unsigned int *) at_b->data;
	     CASE_BODY(new_mpfi_from_u_u, ui_a, ui_b, a, b);
	     break;

	   case SLANG_ULONG_TYPE:
	     ul_b = (unsigned long *) at_b->data;
	     CASE_BODY(new_mpfi_from_u_u, ui_a, ul_b, a, b);
	     break;

	   case SLANG_DOUBLE_TYPE:
	     d_b = (double *) at_b->data;
	     CASE_BODY(new_mpfi_from_u_d, ui_a, d_b, a, b);
	     break;

	   default:		       /* MPFR_Type */
	     fr_b = (MPFR_Type **) at_b->data;
	     CASE_BODY(new_mpfi_from_u_fr, ui_a, fr_b, a, b);
	     break;
	  }
	break;

      case SLANG_ULONG_TYPE:
	ul_a = (unsigned long *)at_a->data;
	switch (at_b->data_type)
	  {
	   case SLANG_INT_TYPE:
	     i_b = (int *) at_b->data;
	     CASE_BODY(new_mpfi_from_u_l, ul_a, i_b, a, b);
	     break;

	   case SLANG_LONG_TYPE:
	     l_b = (long *) at_b->data;
	     CASE_BODY(new_mpfi_from_u_l, ul_a, l_b, a, b);
	     break;

	   case SLANG_UINT_TYPE:
	     ui_b = (unsigned int *) at_b->data;
	     CASE_BODY(new_mpfi_from_u_u, ul_a, ui_b, a, b);
	     break;

	   case SLANG_ULONG_TYPE:
	     ul_b = (unsigned long *) at_b->data;
	     CASE_BODY(new_mpfi_from_u_u, ul_a, ul_b, a, b);
	     break;

	   case SLANG_DOUBLE_TYPE:
	     d_b = (double *) at_b->data;
	     CASE_BODY(new_mpfi_from_u_d, ul_a, d_b, a, b);
	     break;

	   default:		       /* MPFR_Type */
	     fr_b = (MPFR_Type **) at_b->data;
	     CASE_BODY(new_mpfi_from_u_fr, ul_a, fr_b, a, b);
	     break;
	  }
	break;

      case SLANG_DOUBLE_TYPE:
	d_a = (double *)at_a->data;
	switch (at_b->data_type)
	  {
	   case SLANG_INT_TYPE:
	     i_b = (int *) at_b->data;
	     CASE_BODY(new_mpfi_from_d_l, d_a, i_b, a, b);
	     break;

	   case SLANG_LONG_TYPE:
	     l_b = (long *) at_b->data;
	     CASE_BODY(new_mpfi_from_d_l, d_a, l_b, a, b);
	     break;

	   case SLANG_UINT_TYPE:
	     ui_b = (unsigned int *) at_b->data;
	     CASE_BODY(new_mpfi_from_d_u, d_a, ui_b, a, b);
	     break;

	   case SLANG_ULONG_TYPE:
	     ul_b = (unsigned long *) at_b->data;
	     CASE_BODY(new_mpfi_from_d_u, d_a, ul_b, a, b);
	     break;

	   case SLANG_DOUBLE_TYPE:
	     d_b = (double *) at_b->data;
	     CASE_BODY(new_mpfi_from_d_d, d_a, d_b, a, b);
	     break;

	   default:		       /* MPFR_Type */
	     fr_b = (MPFR_Type **) at_b->data;
	     CASE_BODY(new_mpfi_from_d_fr, d_a, fr_b, a, b);
	     break;
	  }
	break;

      default:  /* MPFR_Type */
	fr_a = (MPFR_Type **)at_a->data;
	switch (at_b->data_type)
	  {
	   case SLANG_INT_TYPE:
	     i_b = (int *) at_b->data;
	     CASE_BODY(new_mpfi_from_fr_l, fr_a, i_b, a, b);
	     break;

	   case SLANG_LONG_TYPE:
	     l_b = (long *) at_b->data;
	     CASE_BODY(new_mpfi_from_fr_l, fr_a, l_b, a, b);
	     break;

	   case SLANG_UINT_TYPE:
	     ui_b = (unsigned int *) at_b->data;
	     CASE_BODY(new_mpfi_from_fr_u, fr_a, ui_b, a, b);
	     break;

	   case SLANG_ULONG_TYPE:
	     ul_b = (unsigned long *) at_b->data;
	     CASE_BODY(new_mpfi_from_fr_u, fr_a, ul_b, a, b);
	     break;

	   case SLANG_DOUBLE_TYPE:
	     d_b = (double *) at_b->data;
	     CASE_BODY(new_mpfi_from_fr_d, fr_a, d_b, a, b);
	     break;

	   default:		       /* MPFR_Type */
	     fr_b = (MPFR_Type **) at_b->data;
	     CASE_BODY(new_mpfi_from_fr_fr, fr_a, fr_b, a, b);
	     break;
	  }
	break;
     }
#undef CASE_BODY
   /* drop */

free_and_return:

   mpfr_clear (a);
   mpfr_clear (b);
   return status;
}

/* at_a must be one of the following types:
 *   long, double, mpfr, unsigned long.
 */
static int set_mpfi_array_values_from_scalar (SLang_Array_Type *at_c,
					      SLang_Array_Type *at_a,
					      MPFR_Type *fr_b)
{
   SLuindex_Type i, num;
   MPFI_Type **cp;
   MPFR_Type **fr_a;
   double *d_a;
   unsigned long *ul_a;
   long *l_a;
   int *i_a;
   unsigned int *ui_a;
   mpfr_t a, b;
   int status;

   call_mpfr_init (a);

   num = at_c->num_elements;
   cp = (MPFI_Type **) at_c->data;
   status = -1;

#define CASE_BODY(func,xa,xb,a,b) \
   for (i = 0; i < num; i++) \
   { \
      if (NULL == (cp[i] = func (xa[i], xb, a, b))) \
	goto free_and_return; \
   } status = 0

   switch (at_a->data_type)
     {
      case SLANG_INT_TYPE:
	i_a = (int *)at_a->data;
	CASE_BODY(new_mpfi_from_l_fr, i_a, fr_b, a, b);
	break;

      case SLANG_LONG_TYPE:
	l_a = (long *)at_a->data;
	CASE_BODY(new_mpfi_from_l_fr, l_a, fr_b, a, b);
	break;

      case SLANG_UINT_TYPE:
	ui_a = (unsigned int *)at_a->data;
	CASE_BODY(new_mpfi_from_u_fr, ui_a, fr_b, a, b);
	break;

      case SLANG_ULONG_TYPE:
	ul_a = (unsigned long *)at_a->data;
	CASE_BODY(new_mpfi_from_u_fr, ul_a, fr_b, a, b);
	break;

      case SLANG_DOUBLE_TYPE:
	d_a = (double *)at_a->data;
	CASE_BODY(new_mpfi_from_d_fr, d_a, fr_b, a, b);
	break;

      default:  /* MPFR_Type */
	fr_a = (MPFR_Type **)at_a->data;
	CASE_BODY(new_mpfi_from_fr_fr, fr_a, fr_b, a, b);
	break;
     }
#undef CASE_BODY
   /* drop */

free_and_return:
   mpfr_clear (a);
   return status;
}

static int new_mpfi_array_values (int nargs)
{
   SLang_Array_Type *at_a = NULL, *at_b = NULL, *at_c = NULL;
   MPFR_Type *mpfr_a = NULL, *mpfr_b = NULL;
   int status;

   if (nargs != 2)
     return SLclass_typecast (MPFI_Type_Id, 1, 1);

   switch (SLang_peek_at_stack ())
     {
      case SLANG_ARRAY_TYPE:
	if (-1 == pop_mpfr_array_internal (&at_b))
	  return -1;
	at_c = at_a;
	break;

      default:
	if (-1 == pop_mpfr_type (&mpfr_b))
	  return -1;
     }

   status = -1;
   switch (SLang_peek_at_stack ())
     {
      case SLANG_ARRAY_TYPE:
	if (-1 == pop_mpfr_array_internal (&at_a))
	  goto free_and_return;
	if (at_b != NULL)
	  {
	     if (at_b->num_elements != at_a->num_elements)
	       {
		  SLang_verror (SL_TypeMismatch_Error,
				"Arrays must be of the same size");
		  goto free_and_return;
	       }
	  }
	at_c = at_a;
	break;

      default:
	if (-1 == pop_mpfr_type (&mpfr_a))
	  goto free_and_return;
	break;
     }

   if (at_c == NULL)
     {
	MPFI_Type *mp;

	if (NULL == (mp = new_mpfi_from_fr_fr (mpfr_a, mpfr_b, NULL, NULL)))
	  goto free_and_return;
	status = push_mpfi_type (mp, 1);
	goto free_and_return;
     }

   if (NULL == (at_c = SLang_create_array1 (MPFI_Type_Id, 0, NULL,
					    at_c->dims, at_c->num_dims, 1)))
     goto free_and_return;

   /* 3 cases: (array, scalar), (scalar, array), (array, array) */
   if ((at_a != NULL) && (at_b != NULL))
     status = set_mpfi_array_values_from_arrays (at_c, at_a, at_b);
   else if (at_a != NULL)
     status = set_mpfi_array_values_from_scalar (at_c, at_a, mpfr_b);
   else
     /* The mpfi docs say the order is not important */
     status = set_mpfi_array_values_from_scalar (at_c, at_b, mpfr_a);

   if (status == 0)
     status = SLang_push_array (at_c, 0);

   SLang_free_array (at_c);
   /* drop */

free_and_return:

   if (at_a != NULL)
     SLang_free_array (at_a);
   if (at_b != NULL)
     SLang_free_array (at_b);
   if (mpfr_a != NULL)
     free_mpfr_type (mpfr_a);
   if (mpfr_b != NULL)
     free_mpfr_type (mpfr_b);

   return status;
}

static void mpfi_new_value (void)
{
   MPFI_Type *mp;
   int nargs = SLang_Num_Function_Args;

   if ((SLang_peek_at_stack () == SLANG_ARRAY_TYPE)
       || ((nargs == 2) && (SLANG_ARRAY_TYPE == SLang_peek_at_stack_n(1))))
     {
	(void) new_mpfi_array_values (nargs);
	return;
     }

   if (nargs == 2)
     {
	MPFR_Type *left, *right;

	if (-1 == pop_mpfr_type (&right))
	  return;

	if (-1 == pop_mpfr_type (&left))
	  {
	     free_mpfr_type (right);
	     return;
	  }
	if (NULL == (mp = new_mpfi_type ()))
	  {
	     free_mpfr_type (left);
	     free_mpfr_type (right);
	     return;
	  }
	if (setjmp (LongJmpEnv))
	  {
	     free_mpfr_type (left);
	     free_mpfr_type (right);
	     free_mpfi_type (mp);
	     return;
	  }
	(void) mpfi_interv_fr (mp->value, left->value, right->value);
	free_mpfr_type (left);
	free_mpfr_type (right);
	(void) push_mpfi_type (mp, 1);
	return;
     }

   if (0 == pop_mpfi_type (&mp))
     (void) push_mpfi_type (mp, 1);
}

/*{{{ MPFI Class Methods */

static void mpfi_type_destroy (SLtype type, VOID_STAR v)
{
   MPFI_Type *mp;

   (void) type;
   mp = *(MPFI_Type **) v;
   free_mpfi_type (mp);
}

static int mpfi_type_push (SLtype type, VOID_STAR v)
{
   (void) type;
   return push_mpfi_type (*(MPFI_Type **)v, 0);
}

static int call_mpfi_get_left (mpfr_t left, mpfi_t x)
{
   if (setjmp (LongJmpEnv))
     return -1;

   mpfi_get_left (left, x);
   return 0;
}

static int call_mpfi_get_right (mpfr_t right, mpfi_t x)
{
   if (setjmp (LongJmpEnv))
     return -1;

   mpfi_get_right (right, x);
   return 0;
}

static char *mpfi_type_string (SLtype type, VOID_STAR v)
{
   mpfr_t x;
   char *lstr, *rstr, *str;
   MPFI_Type *mp;
   unsigned int len;

   (void) type;
   mp = *(MPFI_Type **)v;
   if (mp == NULL)
     mp = MPFI_Type_NaN;

   if (-1 == call_mpfr_init (x))
     return NULL;

   if ((-1 == call_mpfi_get_left (x, mp->value))
       || (NULL == (lstr = convert_mpfr_to_string (x))))
     {
	mpfr_clear (x);
	return NULL;
     }

   if ((-1 == call_mpfi_get_right (x, mp->value))
       || (NULL == (rstr = convert_mpfr_to_string (x))))
     {
	mpfr_clear (x);
	SLfree (lstr);
	return NULL;
     }

   mpfr_clear (x);

   /* Need space for "[" + lstr + ", " + rstr + "]" + \0 */
   len = 1 + strlen (lstr) + 2 + strlen (rstr) + 1 + 1;
   if (NULL == (str = SLmalloc (len)))
     {
	SLfree (rstr);
	SLfree (lstr);
	return NULL;
     }
   sprintf (str, "[%s, %s]", lstr, rstr);
   SLfree (rstr);
   SLfree (lstr);
   return str;
}

static void free_mpfi_array (MPFI_Type **mpp, unsigned int n)
{
   unsigned int i;
   for (i = 0; i < n; i++)
     {
	free_mpfi_type (mpp[i]);
	mpp[i] = NULL;
     }
}

static int mpfi_binary_result (int op, SLtype a, SLtype b, SLtype *c)
{
   (void) a; (void) b;

   switch (op)
     {
      default:
	return 0;

      case SLANG_POW:
      case SLANG_PLUS:
      case SLANG_MINUS:
      case SLANG_TIMES:
      case SLANG_DIVIDE:
      case SLANG_MOD:
	*c = MPFI_Type_Id;
	break;

      case SLANG_EQ:
      case SLANG_NE:
      case SLANG_GT:
      case SLANG_GE:
      case SLANG_LT:
      case SLANG_LE:
	*c = SLANG_CHAR_TYPE;
	break;
     }
   return 1;
}

static MPFI_Type *call_mpfi_add (MPFI_Type *a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_add (c->value, a->value, b->value);
   return c;
}

static MPFI_Type *call_mpfi_sub (MPFI_Type *a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_sub (c->value, a->value, b->value);
   return c;
}

static MPFI_Type *call_mpfi_mul (MPFI_Type *a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_mul (c->value, a->value, b->value);
   return c;
}

static MPFI_Type *call_mpfi_div (MPFI_Type *a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_div (c->value, a->value, b->value);
   return c;
}

static int mpfi_mpfi_binary (int op, /*{{{*/
			     SLtype a_type, VOID_STAR ap, unsigned int na,
			     SLtype b_type, VOID_STAR bp, unsigned int nb,
			     VOID_STAR cp)
{
   MPFI_Type **a, **b, **c;
   unsigned int n, n_max;
   unsigned int da, db;
   char *ic;

   (void) a_type;
   (void) b_type;

   a = (MPFI_Type **) ap;
   b = (MPFI_Type **) bp;
   c = (MPFI_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_add (*a, *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_sub (*a, *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_mul (*a, *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_div (*a, *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a, *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfi_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (0 == mpfi_cmp (aa->value, bb->value));
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a, *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfi_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 1;
	     else
	       ic[n] = (0 != mpfi_cmp (aa->value, bb->value));
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a, *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfi_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp (aa->value, bb->value) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a, *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfi_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp (aa->value, bb->value) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a, *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfi_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp (aa->value, bb->value) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a, *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfi_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp (aa->value, bb->value) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	return 0;
     }

   return 1;
}

/*}}}*/

/*{{{ call mpfi with double functions */

static MPFI_Type *call_mpfi_add_d (MPFI_Type *a, double b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_add_d (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_sub_d (MPFI_Type *a, double b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_sub_d (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_mul_d (MPFI_Type *a, double b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_mul_d (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_div_d (MPFI_Type *a, double b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_div_d (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_d_sub (double a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_d_sub (c->value, a, b->value);
   return c;
}

static MPFI_Type *call_mpfi_d_div (double a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_d_div (c->value, a, b->value);
   return c;
}

/*}}}*/

/*{{{ call mpfi with long functions */

static MPFI_Type *call_mpfi_add_si (MPFI_Type *a, long b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_add_si (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_sub_si (MPFI_Type *a, long b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_sub_si (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_mul_si (MPFI_Type *a, long b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_mul_si (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_div_si (MPFI_Type *a, long b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_div_si (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_si_sub (long a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_si_sub (c->value, a, b->value);
   return c;
}

static MPFI_Type *call_mpfi_si_div (long a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_si_div (c->value, a, b->value);
   return c;
}

/*}}}*/

/*{{{ call mpfi with ulong functions */

static MPFI_Type *call_mpfi_add_ui (MPFI_Type *a, unsigned long b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_add_ui (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_sub_ui (MPFI_Type *a, unsigned long b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_sub_ui (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_mul_ui (MPFI_Type *a, unsigned long b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_mul_ui (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_div_ui (MPFI_Type *a, unsigned long b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_div_ui (c->value, a->value, b);
   return c;
}

static MPFI_Type *call_mpfi_ui_sub (unsigned long a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_ui_sub (c->value, a, b->value);
   return c;
}

static MPFI_Type *call_mpfi_ui_div (unsigned long a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_ui_div (c->value, a, b->value);
   return c;
}

/*}}}*/

static MPFI_Type *call_mpfi_add_fr (MPFI_Type *a, MPFR_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_add_fr (c->value, a->value, b->value);
   return c;
}

static MPFI_Type *call_mpfi_sub_fr (MPFI_Type *a, MPFR_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_sub_fr (c->value, a->value, b->value);
   return c;
}

static MPFI_Type *call_mpfi_mul_fr (MPFI_Type *a, MPFR_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;
   if (b == NULL) b = MPFR_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_mul_fr (c->value, a->value, b->value);
   return c;
}

static MPFI_Type *call_mpfi_div_fr (MPFI_Type *a, MPFR_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFI_Type_NaN;
   if (b == NULL) b = MPFR_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_div_fr (c->value, a->value, b->value);
   return c;
}

static MPFI_Type *call_mpfi_fr_sub (MPFR_Type *a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (b == NULL) b = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_fr_sub (c->value, a->value, b->value);
   return c;
}

static MPFI_Type *call_mpfi_fr_div (MPFR_Type *a, MPFI_Type *b)
{
   MPFI_Type *c = new_mpfi_type();
   if (c == NULL) return NULL;
   if (a == NULL) a = MPFR_Type_NaN;
   if (b == NULL) b = MPFI_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (c);
	return NULL;
     }
   mpfi_fr_div (c->value, a->value, b->value);
   return c;
}

static int mpfr_mpfi_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   MPFR_Type **a;
   MPFI_Type **b, **c;
   char *ic;
   unsigned int n, n_max;
   unsigned int da, db;

   (void) a_type;
   (void) b_type;

   a = (MPFR_Type **) ap;
   b = (MPFI_Type **) bp;
   c = (MPFI_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_add_fr (*b, *a)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_fr_sub (*a, *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_mul_fr (*b, *a)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_fr_div (*a, *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a;
	     MPFI_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfr_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (0 == mpfi_cmp_fr (bb->value, aa->value));
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a;
	     MPFI_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfr_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 1;
	     else
	       ic[n] = (0 != mpfi_cmp_fr (bb->value, aa->value));
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a;
	     MPFI_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfr_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_fr (bb->value, aa->value) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a;
	     MPFI_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfr_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_fr (bb->value, aa->value) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a;
	     MPFI_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfr_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_fr (bb->value, aa->value) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFR_Type *aa = *a;
	     MPFI_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || (mpfr_nan_p(aa->value) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_fr (bb->value, aa->value) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	return 0;
     }

   return 1;
}

/*}}}*/

static int mpfi_mpfr_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   MPFI_Type **a, **c;
   MPFR_Type **b;
   char *ic;
   unsigned int n, n_max;
   unsigned int da, db;

   (void) a_type;
   (void) b_type;

   a = (MPFI_Type **) ap;
   b = (MPFR_Type **) bp;
   c = (MPFI_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_add_fr (*a, *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_sub_fr (*a, *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_mul_fr (*a, *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_div_fr (*a, *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     MPFR_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || mpfi_nan_p(aa->value) || mpfr_nan_p(bb->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_fr (aa->value, bb->value) == 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     MPFR_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || mpfi_nan_p(aa->value) || mpfr_nan_p(bb->value))
	       ic[n] = 1;
	     else
	       ic[n] = (mpfi_cmp_fr (aa->value, bb->value) != 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     MPFR_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || mpfi_nan_p(aa->value) || mpfr_nan_p(bb->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_fr (aa->value, bb->value) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     MPFR_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || mpfi_nan_p(aa->value) || mpfr_nan_p(bb->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_fr (aa->value, bb->value) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     MPFR_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || mpfi_nan_p(aa->value) || mpfr_nan_p(bb->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_fr (aa->value, bb->value) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     MPFR_Type *bb = *b;
	     if ((aa == NULL) || (bb == NULL)
		 || mpfi_nan_p(aa->value) || mpfr_nan_p(bb->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_fr (aa->value, bb->value) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	return 0;
     }

   return 1;
}

/*}}}*/

static int double_mpfi_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   char *a;
   MPFI_Type **b, **c;
   char *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   SLang_To_Double_Fun_Type to_double;
   unsigned int sizeof_a;

   if (NULL == (to_double = SLarith_get_to_double_fun (a_type, &sizeof_a)))
     return 0;

   (void) b_type;

   a = (char *) ap;
   b = (MPFI_Type **) bp;
   c = (MPFI_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = sizeof_a;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_add_d (*b, to_double((VOID_STAR)a))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_d_sub (to_double((VOID_STAR)a), *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_mul_d (*b, to_double((VOID_STAR)a))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_d_div (to_double((VOID_STAR)a), *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double aa = to_double((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (isnan(aa) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (0 == mpfi_cmp_d (bb->value, aa));
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double aa = to_double((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (isnan(aa) || (mpfi_nan_p(bb->value))))
	       ic[n] = 1;
	     else
	       ic[n] = (0 != mpfi_cmp_d (bb->value, aa));
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double aa = to_double((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (isnan(aa) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_d (bb->value, aa) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double aa = to_double((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (isnan(aa) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_d (bb->value, aa) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double aa = to_double((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (isnan(aa) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_d (bb->value, aa) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     double aa = to_double((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (isnan(aa) || (mpfi_nan_p(bb->value))))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_d (bb->value, aa) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	return 0;
     }

   return 1;
}

/*}}}*/

static int mpfi_double_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   MPFI_Type **a, **c;
   char *b, *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   SLang_To_Double_Fun_Type to_double;
   unsigned int sizeof_b;

   if (NULL == (to_double = SLarith_get_to_double_fun (b_type, &sizeof_b)))
     return 0;

   (void) a_type;

   a = (MPFI_Type **) ap;
   b = (char *) bp;
   c = (MPFI_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = sizeof_b;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_add_d (*a, to_double((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_sub_d (*a, to_double((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_mul_d (*a, to_double((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_div_d (*a, to_double((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     double bb = (*to_double)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) || isnan(bb))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_d (aa->value, bb) == 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     double bb = (*to_double)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) || isnan(bb))
	       ic[n] = 1;
	     else
	       ic[n] = (mpfi_cmp_d (aa->value, bb) != 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     double bb = (*to_double)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) || isnan(bb))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_d (aa->value, bb) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     double bb = (*to_double)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) || isnan(bb))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_d (aa->value, bb) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     double bb = (*to_double)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) || isnan(bb))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_d (aa->value, bb) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     double bb = (*to_double)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) || isnan(bb))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_d (aa->value, bb) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	return 0;
     }

   return 1;
}

/*}}}*/

static int long_mpfi_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   char *a;
   MPFI_Type **b, **c;
   char *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   To_Long_Fun_Type to_long;
   unsigned int sizeof_a;

   if (NULL == (to_long = get_to_long_fun (a_type, &sizeof_a)))
     return 0;

   (void) b_type;

   a = (char *) ap;
   b = (MPFI_Type **) bp;
   c = (MPFI_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = sizeof_a;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_add_si (*b, to_long((VOID_STAR)a))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_si_sub (to_long((VOID_STAR)a), *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_mul_si (*b, to_long((VOID_STAR)a))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_si_div (to_long((VOID_STAR)a), *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long aa = to_long((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (mpfi_nan_p(bb->value)))
	       ic[n] = 0;
	     else
	       ic[n] = (0 == mpfi_cmp_si (bb->value, aa));
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long aa = to_long((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (mpfi_nan_p(bb->value)))
	       ic[n] = 1;
	     else
	       ic[n] = (0 != mpfi_cmp_si (bb->value, aa));
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long aa = to_long((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (mpfi_nan_p(bb->value)))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_si (bb->value, aa) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long aa = to_long((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (mpfi_nan_p(bb->value)))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_si (bb->value, aa) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long aa = to_long((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (mpfi_nan_p(bb->value)))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_si (bb->value, aa) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     long aa = to_long((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || (mpfi_nan_p(bb->value)))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_si (bb->value, aa) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	return 0;
     }

   return 1;
}

/*}}}*/

static int mpfi_long_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   MPFI_Type **a, **c;
   char *b, *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   To_Long_Fun_Type to_long;
   unsigned int sizeof_b;

   if (NULL == (to_long = get_to_long_fun (b_type, &sizeof_b)))
     return 0;

   (void) a_type;

   a = (MPFI_Type **) ap;
   b = (char *) bp;
   c = (MPFI_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = sizeof_b;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_add_si (*a, to_long((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_sub_si (*a, to_long((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_mul_si (*a, to_long((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_div_si (*a, to_long((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     long bb = (*to_long)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_si (aa->value, bb) == 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     long bb = (*to_long)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value))
	       ic[n] = 1;
	     else
	       ic[n] = (mpfi_cmp_si (aa->value, bb) != 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     long bb = (*to_long)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_si (aa->value, bb) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     long bb = (*to_long)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_si (aa->value, bb) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     long bb = (*to_long)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_si (aa->value, bb) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     long bb = (*to_long)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_si (aa->value, bb) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	return 0;
     }

   return 1;
}

/*}}}*/

static int ulong_mpfi_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   char *a;
   MPFI_Type **b, **c;
   char *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   To_ULong_Fun_Type to_ulong;
   unsigned int sizeof_a;

   if (NULL == (to_ulong = get_to_ulong_fun (a_type, &sizeof_a)))
     return 0;

   (void) b_type;

   a = (char *) ap;
   b = (MPFI_Type **) bp;
   c = (MPFI_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = sizeof_a;
   if (nb == 1) db = 0; else db = 1;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_add_ui (*b, to_ulong((VOID_STAR)a))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_ui_sub (to_ulong((VOID_STAR)a), *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_mul_ui (*b, to_ulong((VOID_STAR)a))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_ui_div (to_ulong((VOID_STAR)a), *b)))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long aa = to_ulong((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || mpfi_nan_p(bb->value))
	       ic[n] = 0;
	     else
	       ic[n] = (0 == mpfi_cmp_ui (bb->value, aa));
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long aa = to_ulong((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || mpfi_nan_p(bb->value))
	       ic[n] = 1;
	     else
	       ic[n] = (0 != mpfi_cmp_ui (bb->value, aa));
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long aa = to_ulong((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || mpfi_nan_p(bb->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_ui (bb->value, aa) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long aa = to_ulong((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || mpfi_nan_p(bb->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_ui (bb->value, aa) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long aa = to_ulong((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || mpfi_nan_p(bb->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_ui (bb->value, aa) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     unsigned long aa = to_ulong((VOID_STAR)a);
	     MPFI_Type *bb = *b;
	     if ((bb == NULL)
		 || mpfi_nan_p(bb->value))
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_ui (bb->value, aa) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	return 0;
     }

   return 1;
}

/*}}}*/

static int mpfi_ulong_binary (int op, /*{{{*/
				SLtype a_type, VOID_STAR ap, unsigned int na,
				SLtype b_type, VOID_STAR bp, unsigned int nb,
				VOID_STAR cp)
{
   MPFI_Type **a, **c;
   char *b, *ic;
   unsigned int n, n_max;
   unsigned int da, db;
   To_ULong_Fun_Type to_ulong;
   unsigned int sizeof_b;

   if (NULL == (to_ulong = get_to_ulong_fun (b_type, &sizeof_b)))
     return 0;

   (void) a_type;

   a = (MPFI_Type **) ap;
   b = (char *) bp;
   c = (MPFI_Type **) cp;
   ic = (char *) cp;

   if (na == 1) da = 0; else da = 1;
   if (nb == 1) db = 0; else db = sizeof_b;

   if (na > nb) n_max = na; else n_max = nb;

   switch (op)
     {
      default:
	return 0;

      case SLANG_PLUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_add_ui (*a, to_ulong((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_MINUS:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_sub_ui (*a, to_ulong((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_TIMES:
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_mul_ui (*a, to_ulong((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_DIVIDE:	       /* / */
	for (n = 0; n < n_max; n++)
	  {
	     if (NULL == (c[n] = call_mpfi_div_ui (*a, to_ulong((VOID_STAR)b))))
	       {
		  free_mpfi_array (c, n);
		  return -1;
	       }
	     a += da; b += db;
	  }
	break;

      case SLANG_EQ:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     unsigned long bb = (*to_ulong)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) )
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_ui (aa->value, bb) == 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_NE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     unsigned long bb = (*to_ulong)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) )
	       ic[n] = 1;
	     else
	       ic[n] = (mpfi_cmp_ui (aa->value, bb) != 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     unsigned long bb = (*to_ulong)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) )
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_ui (aa->value, bb) > 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_GE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     unsigned long bb = (*to_ulong)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) )
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_ui (aa->value, bb) >= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LT:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     unsigned long bb = (*to_ulong)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) )
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_ui (aa->value, bb) < 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_LE:
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < n_max; n++)
	  {
	     MPFI_Type *aa = *a;
	     unsigned long bb = (*to_ulong)((VOID_STAR)b);
	     if ((aa == NULL)
		 || mpfi_nan_p(aa->value) )
	       ic[n] = 0;
	     else
	       ic[n] = (mpfi_cmp_ui (aa->value, bb) <= 0);
	     a += da; b += db;
	  }
	break;

      case SLANG_POW:
	return 0;
     }

   return 1;
}

/*}}}*/

/*{{{ MPFI Unary Functions */

static int mpfi_unary_result (int op, SLtype a, SLtype *b)
{
   (void) a;

   switch (op)
     {
      case SLANG_PLUSPLUS:
      case SLANG_MINUSMINUS:
      case SLANG_SIGN:
      default:
	return 0;

      case SLANG_CHS:
      case SLANG_MUL2:
      case SLANG_SQR:
      case SLANG_ABS:
	*b = MPFI_Type_Id;
	break;

      case SLANG_ISPOS:
      case SLANG_ISNEG:
      case SLANG_ISNONNEG:
	*b = SLANG_CHAR_TYPE;
	break;
     }
   return 1;
}

static MPFI_Type *call_mpfi_neg (MPFI_Type *a)
{
   MPFI_Type *b = new_mpfi_type();
   if (b == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (b);
	return NULL;
     }
   (void) mpfi_neg (b->value, a->value);
   return b;
}

static MPFI_Type *call_mpfi_sqr (MPFI_Type *a)
{
   MPFI_Type *b = new_mpfi_type();
   if (b == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (b);
	return NULL;
     }
   (void) mpfi_sqr (b->value, a->value);
   return b;
}

static MPFI_Type *call_mpfi_abs (MPFI_Type *a)
{
   MPFI_Type *b = new_mpfi_type();
   if (b == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (b);
	return NULL;
     }
   (void) mpfi_abs (b->value, a->value);
   return b;
}

static int mpfi_unary (int op,
		       SLtype a_type, VOID_STAR ap, unsigned int na,
		       VOID_STAR bp)
{
   unsigned int n;
   MPFI_Type **a, **b;
   char *c;
   MPFI_Type *two;

   (void) a_type;
   a = (MPFI_Type **) ap;
   b = (MPFI_Type **) bp;

   switch (op)
     {
      case SLANG_PLUSPLUS:
      case SLANG_MINUSMINUS:
      case SLANG_SIGN:
      default:
	return 0;

      case SLANG_CHS:
	for (n = 0; n < na; n++)
	  {
	     if (NULL == (b[n] = call_mpfi_neg (a[n])))
	       {
		  free_mpfi_array (b, n);
		  return -1;
	       }
	  }
	break;

      case SLANG_SQR:
	for (n = 0; n < na; n++)
	  {
	     if (NULL == (b[n] = call_mpfi_sqr (a[n])))
	       {
		  free_mpfi_array (b, n);
		  return -1;
	       }
	  }
	break;

      case SLANG_MUL2:
	if (NULL == (two = new_mpfi_type_from_double (2.0)))
	  return -1;

	for (n = 0; n < na; n++)
	  {
	     if (NULL == (b[n] = call_mpfi_mul (two, a[n])))
	       {
		  free_mpfi_type (two);
		  free_mpfi_array (b, n);
		  return -1;
	       }
	  }
	free_mpfi_type (two);
	break;

      case SLANG_ABS:
	for (n = 0; n < na; n++)
	  {
	     if (NULL == (b[n] = call_mpfi_abs (a[n])))
	       {
		  free_mpfi_array (b, n);
		  return -1;
	       }
	  }
	break;

      case SLANG_ISPOS:
	c = (char *) bp;
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < na; n++)
	  {
	     if (b[n] == NULL)
	       c[n] = 0;
	     else
	       c[n] = (mpfi_is_pos(b[n]->value) > 0);
	  }
	break;

      case SLANG_ISNEG:
	c = (char *) bp;
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < na; n++)
	  {
	     if (b[n] == NULL)
	       c[n] = 0;
	     else
	       c[n] = (mpfi_is_neg(b[n]->value) > 0);
	  }
	break;

      case SLANG_ISNONNEG:
	c = (char *) bp;
	if (setjmp (LongJmpEnv))
	  return -1;
	for (n = 0; n < na; n++)
	  {
	     if (b[n] == NULL)
	       c[n] = 0;
	     else
	       c[n] = (mpfi_is_nonneg(b[n]->value) > 0);
	  }
	break;
     }
   return 1;
}

/*}}}*/

/*{{{ MPFI Math Unary Functions */

static MPFI_Type *call_mpfi_int_func2 (int (*func)(mpfi_ptr, mpfi_srcptr), MPFI_Type *a)
{
   MPFI_Type *b = new_mpfi_type();
   if (b == NULL)
     return NULL;
   if (a == NULL) a = MPFI_Type_NaN;
   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (b);
	return NULL;
     }
   (void) (*func) (b->value, a->value);
   return b;
}

static int mpfi_math_unary_result (int op, SLtype a, SLtype *b) /*{{{*/
{
   (void) a;

   switch (op)
     {
      case SLMATH_SINH:
      case SLMATH_COSH:
      case SLMATH_TANH:
      case SLMATH_TAN:
      case SLMATH_ASIN:
      case SLMATH_ACOS:
      case SLMATH_ATAN:
      case SLMATH_EXP:
      case SLMATH_LOG:
      case SLMATH_LOG10:
      case SLMATH_SQRT:
      case SLMATH_SIN:
      case SLMATH_COS:
      case SLMATH_ASINH:
      case SLMATH_ATANH:
      case SLMATH_ACOSH:
      case SLMATH_EXPM1:
      case SLMATH_LOG1P:
	*b = MPFI_Type_Id;
	break;

      case SLMATH_ISINF:
      case SLMATH_ISNAN:
	*b = SLANG_CHAR_TYPE;
	break;

      case SLMATH_FLOOR:
      case SLMATH_CEIL:
      case SLMATH_ROUND:
      default:
	return 0;
     }

   return 1;
}

/*}}}*/

static int mpfi_math_unary (int op, /*{{{*/
			    SLtype a_type, VOID_STAR ap, unsigned int na,
			    VOID_STAR bp)
{
   unsigned int i;
   MPFI_Type **a, **b;
   int (*func)(mpfi_ptr, mpfi_srcptr);

   (void) a_type;
   a = (MPFI_Type **) ap;
   b = (MPFI_Type **) bp;

   switch (op)
     {
      case SLMATH_SINH: func = mpfi_sinh; break;
      case SLMATH_COSH: func = mpfi_cosh; break;
      case SLMATH_TANH: func = mpfi_tanh; break;
      case SLMATH_TAN: func = mpfi_tan; break;
      case SLMATH_ASIN: func = mpfi_asin; break;
      case SLMATH_ACOS: func = mpfi_acos; break;
      case SLMATH_ATAN: func = mpfi_atan; break;
      case SLMATH_EXP: func = mpfi_exp; break;
      case SLMATH_LOG: func = mpfi_log; break;
      case SLMATH_LOG10: func = mpfi_log10; break;
      case SLMATH_SQRT: func = mpfi_sqrt; break;
      case SLMATH_SIN: func = mpfi_sin; break;
      case SLMATH_COS: func = mpfi_cos; break;
      case SLMATH_ASINH: func = mpfi_asinh; break;
      case SLMATH_ATANH: func = mpfi_atanh; break;
      case SLMATH_ACOSH: func = mpfi_acosh; break;
      case SLMATH_EXPM1: func = mpfi_expm1; break;
      case SLMATH_LOG1P: func = mpfi_log1p; break;

      case SLMATH_ISINF:
	  {
	     char *c = (char *)bp;
	     if (setjmp (LongJmpEnv))
	       return -1;
	     for (i = 0; i < na; i++)
	       {
		  if (a[i] == NULL)
		    c[i] = 0;
		  else
		    c[i] = (char) mpfi_inf_p (a[i]->value);
	       }
	  }
	return 1;

      case SLMATH_ISNAN:
	  {
	     char *c = (char *)bp;
	     if (setjmp (LongJmpEnv))
	       return -1;
	     for (i = 0; i < na; i++)
	       {
		  if (a[i] == NULL)
		    c[i] = 1;
		  else
		    c[i] = (char) mpfi_nan_p (a[i]->value);
	       }
	  }
	return 1;

      case SLMATH_FLOOR:
      case SLMATH_CEIL:
      case SLMATH_ROUND:
      default:
	return 0;
     }

   for (i = 0; i < na; i++)
     {
	if (NULL == (b[i] = call_mpfi_int_func2 (func, a[i])))
	  {
	     free_mpfi_array (b, i);
	     return -1;
	  }
     }

   return 1;
}

/*}}}*/

/*}}}*/

#define TYPECAST_BODY(type, func) \
   { \
      type *ptr = (type *) from; \
      for (n = 0; n < num; n++) \
	{ \
	   MPFI_Type *mp; \
	   type x = *ptr++; \
	   if (NULL == (mp = func (x))) \
	     { \
		free_mpfi_array (mpp, n); \
		return -1; \
	     } \
	   *mpp++ = mp; \
	} \
   }

static int mpfi_typecast (SLtype from_type, VOID_STAR from, unsigned int num, /*{{{*/
	       SLtype to_type, VOID_STAR to)
{
   unsigned int n;
   MPFI_Type **mpp;

   (void) to_type;

   mpp = (MPFI_Type **) to;
   memset ((char *)mpp, 0, num*sizeof(MPFI_Type *));

   switch (from_type)
     {
      case SLANG_UCHAR_TYPE:
	TYPECAST_BODY(unsigned char, new_mpfi_type_from_ulong) return 1;
      case SLANG_USHORT_TYPE:
	TYPECAST_BODY(unsigned short, new_mpfi_type_from_ulong) return 1;
      case SLANG_UINT_TYPE:
	TYPECAST_BODY(unsigned int, new_mpfi_type_from_ulong) return 1;
      case SLANG_ULONG_TYPE:
	TYPECAST_BODY(unsigned long, new_mpfi_type_from_ulong) return 1;

      case SLANG_CHAR_TYPE:
	TYPECAST_BODY(char, new_mpfi_type_from_long) return 1;
      case SLANG_SHORT_TYPE:
	TYPECAST_BODY(short, new_mpfi_type_from_long) return 1;
      case SLANG_INT_TYPE:
	TYPECAST_BODY(int, new_mpfi_type_from_long) return 1;
      case SLANG_LONG_TYPE:
	TYPECAST_BODY(long, new_mpfi_type_from_long) return 1;

      case SLANG_FLOAT_TYPE:
	TYPECAST_BODY(float, new_mpfi_type_from_double) return 1;
      case SLANG_DOUBLE_TYPE:
	TYPECAST_BODY(double, new_mpfi_type_from_double) return 1;

      case SLANG_STRING_TYPE:
	TYPECAST_BODY(char *, new_mpfi_type_from_string) return 1;

      case SLANG_BSTRING_TYPE:
	  {
	     SLang_BString_Type **ptr;

	     ptr = (SLang_BString_Type **) from;

	     for (n = 0; n < num; n++)
	       {
		  MPFI_Type *mp;
		  char *s;
		  unsigned int len;

		  s = (char *)SLbstring_get_pointer (*ptr, &len);
		  if (NULL == (mp = new_mpfi_type_from_string (s)))
		    {
		       free_mpfi_array (mpp, n);
		       return -1;
		    }
		  *mpp++ = mp;
		  ptr++;
	       }
	  }
	return 1;

      default:
	if (from_type == (SLtype) MPFR_Type_Id)
	  {
	     TYPECAST_BODY(MPFR_Type *, new_mpfi_type_from_mpfr_type) return 1;
	  }
	break;
     }
   return 0;
}
#undef TYPECAST_BODY

/*}}}*/

static int init_mpfi_class (void)
{
   SLang_Class_Type *cl;
   unsigned int i;

   if (MPFI_Type_Id != -1)
     return 0;

   if (NULL == (cl = SLclass_allocate_class ("MPFI_Type")))
     return -1;

   (void) SLclass_set_string_function (cl, mpfi_type_string);
   (void) SLclass_set_push_function (cl, mpfi_type_push);
   (void) SLclass_set_destroy_function (cl, mpfi_type_destroy);

   if (-1 == SLclass_register_class (cl, SLANG_VOID_TYPE, sizeof (VOID_STAR),
				     SLANG_CLASS_TYPE_PTR))
     return -1;

   MPFI_Type_Id = SLclass_get_class_id (cl);

   for (i = 0; i < NUM_INT_TYPES; i++)
     {
	SLtype t = Integer_Types[i];
	if (-1 == (SLclass_add_typecast(t, MPFI_Type_Id, mpfi_typecast, 1)))
	  return -1;
	if (-1 == SLclass_add_binary_op (t, MPFI_Type_Id, long_mpfi_binary, mpfi_binary_result))
	  return -1;
	if (-1 == SLclass_add_binary_op (MPFI_Type_Id, t, mpfi_long_binary, mpfi_binary_result))
	  return -1;
     }

   for (i = 0; i < NUM_UINT_TYPES; i++)
     {
	SLtype t = UInteger_Types[i];
	if (-1 == (SLclass_add_typecast(t, MPFI_Type_Id, mpfi_typecast, 1)))
	  return -1;
	if (-1 == SLclass_add_binary_op (t, MPFI_Type_Id, ulong_mpfi_binary, mpfi_binary_result))
	  return -1;
	if (-1 == SLclass_add_binary_op (MPFI_Type_Id, t, mpfi_ulong_binary, mpfi_binary_result))
	  return -1;
     }

   for (i = 0; i < NUM_FLOAT_TYPES; i++)
     {
	SLtype t = Float_Types[i];
	if (-1 == (SLclass_add_typecast(t, MPFI_Type_Id, mpfi_typecast, 1)))
	  return -1;

	if (-1 == SLclass_add_binary_op (t, MPFI_Type_Id, double_mpfi_binary, mpfi_binary_result))
	  return -1;
	if (-1 == SLclass_add_binary_op (MPFI_Type_Id, t, mpfi_double_binary, mpfi_binary_result))
	  return -1;
     }

   if (-1 == SLclass_add_typecast (MPFR_Type_Id, MPFI_Type_Id, mpfi_typecast, 1))
     return -1;

   if ((-1 == SLclass_add_binary_op (MPFI_Type_Id, MPFR_Type_Id, mpfi_mpfr_binary, mpfi_binary_result))
       || (-1 == SLclass_add_binary_op (MPFR_Type_Id, MPFI_Type_Id, mpfr_mpfi_binary, mpfi_binary_result)))
     return -1;

#if 1
   if ((-1 == SLclass_add_typecast (SLANG_STRING_TYPE, MPFI_Type_Id, mpfi_typecast, 1))
       || (-1 == SLclass_add_typecast (SLANG_BSTRING_TYPE, MPFI_Type_Id, mpfi_typecast, 1)))
     return -1;
#endif
   if (-1 == SLclass_add_binary_op (MPFI_Type_Id, MPFI_Type_Id, mpfi_mpfi_binary, mpfi_binary_result))
     return -1;

   if (-1 == (SLclass_add_unary_op (MPFI_Type_Id, mpfi_unary, mpfi_unary_result)))
     return -1;

   if (-1 == SLclass_add_math_op (MPFI_Type_Id, mpfi_math_unary, mpfi_math_unary_result))
     return -1;

   slMPFI_Error = SLerr_new_exception (SL_RunTime_Error, "MPFI_Error", "MPFI Error");

   if (NULL == (MPFI_Type_NaN = new_mpfi_type ()))
     return -1;

   return 0;
}

/*}}}*/

static void not_yet_implemented_array (void)
{
   SLang_verror (SL_NotImplemented_Error, "This function does not yet support arrays");
}

static int get_mpfi_endpoints (MPFI_Type *mp, MPFR_Type **leftp, MPFR_Type **rightp)
{
   MPFR_Type *left, *right;

   *leftp = *rightp = NULL;

   if (NULL == (left = new_mpfr_type ()))
     return -1;
   if (NULL == (right = new_mpfr_type ()))
     {
	free_mpfr_type (left);
	return -1;
     }
   if (mp == NULL)
     mp = MPFI_Type_NaN;

   if ((-1 == call_mpfi_get_left (left->value, mp->value))
       || (-1 == call_mpfi_get_right (right->value, mp->value)))
     {
	free_mpfr_type (left);
	free_mpfr_type (right);
	return -1;
     }
   *leftp = left;
   *rightp = right;
   return 0;
}

static void intrin_mpfi_get_leftright (void)
{
   MPFI_Type *mp, **fip;
   MPFR_Type *left, *right, **leftp, **rightp;
   SLang_Array_Type *at;
   SLang_Array_Type *at_left, *at_right;
   SLuindex_Type i, num;

   if (SLang_peek_at_stack () != SLANG_ARRAY_TYPE)
     {
	if (-1 == pop_mpfi_type (&mp))
	  return;

	if (0 == get_mpfi_endpoints (mp, &left, &right))
	  {
	     (void) push_mpfr_type (left, 1);
	     (void) push_mpfr_type (right, 1);
	  }

	free_mpfi_type (mp);
	return;
     }
   if (-1 == SLang_pop_array_of_type (&at, MPFI_Type_Id))
     return;
   if (NULL == (at_left = SLang_create_array1 (MPFR_Type_Id, 0, NULL, at->dims, at->num_dims, 1)))
     {
	SLang_free_array (at);
	return;
     }

   if (NULL == (at_right = SLang_create_array1 (MPFR_Type_Id, 0, NULL, at->dims, at->num_dims, 1)))
     {
	SLang_free_array (at_left);
	SLang_free_array (at);
	return;
     }

   num = at->num_elements;
   fip = (MPFI_Type **) at->data;
   leftp = (MPFR_Type **) at_left->data;
   rightp = (MPFR_Type **) at_right->data;
   for (i = 0; i < num; i++)
     {
	if (-1 == get_mpfi_endpoints (fip[i], &left, &right))
	  goto free_return;

	leftp[i] = left;
	rightp[i] = right;
     }
   (void) SLang_push_array (at_left, 0);
   (void) SLang_push_array (at_right, 0);
   /* drop */
free_return:
   SLang_free_array (at_right);
   SLang_free_array (at_left);
   SLang_free_array (at);
}

static void intrin_mpfi_put (void)
{
   MPFI_Type *fi0, *fi1;

   if (SLang_peek_at_stack_n(1) == SLANG_ARRAY_TYPE)
     {
	not_yet_implemented_array ();
	return;
     }

   if (-1 == pop_mpfi_type (&fi1))
     return;
   if (-1 == pop_mpfi_type (&fi0))
     {
	free_mpfi_type (fi1);
	return;
     }
   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (fi0);
	free_mpfi_type (fi1);
	return;
     }

   (void) mpfi_put (fi0->value, fi1->value);
   free_mpfi_type (fi0);
   free_mpfi_type (fi1);
}

static int do_intrin_mpfi_func0 (int (*func)(mpfi_t))
{
   MPFI_Type *fi;

   if (NULL == (fi = new_mpfi_type ()))
     return -1;

   if (setjmp (LongJmpEnv))
     {
	free_mpfi_type (fi);
	return -1;
     }
   (void) (*func)(fi->value);

   return push_mpfi_type (fi, 1);
}

static void intrin_mpfi_const_pi (void)
{
   (void) do_intrin_mpfi_func0 (mpfi_const_pi);
}

static void intrin_mpfi_const_log2 (void)
{
   (void) do_intrin_mpfi_func0 (mpfi_const_log2);
}

/*}}}*/

static SLang_Intrin_Fun_Type Module_Intrinsics [] =
{
   MAKE_INTRINSIC_0("mpfi", mpfi_new_value, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("mpfi_put", intrin_mpfi_put, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("mpfi_const_pi", intrin_mpfi_const_pi, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("mpfi_const_log2", intrin_mpfi_const_log2, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("mpfr_const_pi", intrin_mpfr_const_pi, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("mpfr_const_log2", intrin_mpfr_const_log2, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("mpfr", mpfr_new_value, SLANG_VOID_TYPE),
   MAKE_INTRINSIC_0("mpfr_get_default_prec", intrin_mpfr_get_def_prec, SLANG_ULONG_TYPE),
   MAKE_INTRINSIC_1("mpfr_set_default_prec", intrin_mpfr_set_def_prec, SLANG_VOID_TYPE, SLANG_ULONG_TYPE),
   MAKE_INTRINSIC_0("mpfr_get_prec", intrin_mpfr_get_prec, SLANG_ULONG_TYPE),

   MAKE_INTRINSIC_0("mpfi_interval", intrin_mpfi_get_leftright, SLANG_VOID_TYPE),
   SLANG_END_INTRIN_FUN_TABLE
};

static char *Module_Version_String = MODULE_VERSION_STRING;
static SLang_Intrin_Var_Type Module_Variables [] =
{
   MAKE_VARIABLE("_mpfi_module_version_string", &Module_Version_String, SLANG_STRING_TYPE, 1),
   SLANG_END_INTRIN_VAR_TABLE
};

static SLang_IConstant_Type Module_IConstants [] =
{
   MAKE_ICONSTANT("_mpfi_module_version", MODULE_VERSION_NUMBER),
   SLANG_END_ICONST_TABLE
};

static SLang_DConstant_Type Module_DConstants [] =
{
   SLANG_END_DCONST_TABLE
};

static int pre_init (void)
{
   static int inited = 0;

   if (inited)
     return 0;

   mp_set_memory_functions (&mem_malloc, &mem_realloc, &mem_free);
   Double_NaN = log(-1.0);

   if (-1 == init_mpfr_class ())
     return -1;

   if (-1 == init_mpfi_class ())
     return -1;

   inited = 1;
   return 0;
}

static unsigned long Min_MPFR_Prec = MPFR_PREC_MIN;
static unsigned long Max_MPFR_Prec = MPFR_PREC_MAX;

int init_mpfp_module_ns (char *ns_name)
{
   SLang_NameSpace_Type *ns = SLns_create_namespace (ns_name);
   if (ns == NULL)
     return -1;

   if ((-1 == pre_init ())
       || (-1 == SLns_add_intrin_var_table (ns, Module_Variables, NULL))
       || (-1 == SLns_add_intrin_fun_table (ns, Module_Intrinsics, NULL))
       || (-1 == SLns_add_iconstant_table (ns, Module_IConstants, NULL))
       || (-1 == SLns_add_dconstant_table (ns, Module_DConstants, NULL))
       || (-1 == SLns_add_intrinsic_variable (ns, "MPFR_PREC_MIN", (VOID_STAR)&Min_MPFR_Prec, SLANG_ULONG_TYPE, 1))
       || (-1 == SLns_add_intrinsic_variable (ns, "MPFR_PREC_MAX", (VOID_STAR)&Max_MPFR_Prec, SLANG_ULONG_TYPE, 1))
      )
     return -1;

   return 0;
}

/* This function is optional */
void deinit_mpfp_module (void)
{
   free_mpfr_type (MPFR_Type_NaN);
   free_mpfi_type (MPFI_Type_NaN);
   mpfr_free_cache ();
}
